from unittest.mock import AsyncMock

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement, SwitchbotModel
from switchbot.devices import roller_shade

from .test_adv_parser import generate_ble_device


def create_device_for_command_testing(
    position=50, calibration=True, reverse_mode=False
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    roller_shade_device = roller_shade.SwitchbotRollerShade(
        ble_device, reverse_mode=reverse_mode
    )
    roller_shade_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, position, calibration)
    )
    roller_shade_device._send_multiple_commands = AsyncMock()
    roller_shade_device.update = AsyncMock()
    return roller_shade_device


def make_advertisement_data(
    ble_device: BLEDevice, in_motion: bool, position: int, calibration: bool = True
):
    """Set advertisement data with defaults."""
    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": b",\x00'\x9f\x11\x04",
            "data": {
                "battery": 39,
                "calibration": calibration,
                "deviceChain": 1,
                "inMotion": in_motion,
                "lightLevel": 1,
                "position": position,
            },
            "isEncrypted": False,
            "model": ",",
            "modelFriendlyName": "Roller Shade",
            "modelName": SwitchbotModel.ROLLER_SHADE,
        },
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
async def test_open():
    roller_shade_device = create_device_for_command_testing()
    await roller_shade_device.open()
    assert roller_shade_device.is_opening() is True
    assert roller_shade_device.is_closing() is False
    roller_shade_device._send_multiple_commands.assert_awaited_once_with(
        [roller_shade.OPEN_KEYS[0], f"{roller_shade.OPEN_KEYS[1]}0000"]
    )


@pytest.mark.asyncio
async def test_open_quietdrift():
    roller_shade_device = create_device_for_command_testing()
    await roller_shade_device.open(mode=1)
    assert roller_shade_device.is_opening() is True
    assert roller_shade_device.is_closing() is False
    roller_shade_device._send_multiple_commands.assert_awaited_once_with(
        [roller_shade.OPEN_KEYS[0], f"{roller_shade.OPEN_KEYS[1]}0100"]
    )


@pytest.mark.asyncio
async def test_close():
    roller_shade_device = create_device_for_command_testing()
    await roller_shade_device.close()
    assert roller_shade_device.is_opening() is False
    assert roller_shade_device.is_closing() is True
    roller_shade_device._send_multiple_commands.assert_awaited_once_with(
        [roller_shade.CLOSE_KEYS[0], f"{roller_shade.CLOSE_KEYS[1]}0064"]
    )


@pytest.mark.asyncio
async def test_close_quietdrift():
    roller_shade_device = create_device_for_command_testing()
    await roller_shade_device.close(mode=1)
    assert roller_shade_device.is_opening() is False
    assert roller_shade_device.is_closing() is True
    roller_shade_device._send_multiple_commands.assert_awaited_once_with(
        [roller_shade.CLOSE_KEYS[0], f"{roller_shade.CLOSE_KEYS[1]}0164"]
    )


@pytest.mark.asyncio
async def test_get_basic_info_returns_none_when_no_data():
    roller_shade_device = create_device_for_command_testing()
    roller_shade_device._get_basic_info = AsyncMock(return_value=None)

    assert await roller_shade_device.get_basic_info() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("reverse_mode", "data", "result"),
    [
        (
            True,
            bytes([0, 1, 10, 2, 0, 50, 4]),
            [1, 1, 2, "anticlockwise", False, False, False, False, False, 50, 4],
        ),
        (
            True,
            bytes([0, 1, 10, 2, 214, 50, 4]),
            [1, 1, 2, "clockwise", True, False, True, True, True, 50, 4],
        ),
    ],
)
async def test_get_basic_info(reverse_mode, data, result):
    blind_device = create_device_for_command_testing(reverse_mode=reverse_mode)
    blind_device._get_basic_info = AsyncMock(return_value=data)

    info = await blind_device.get_basic_info()
    assert info["battery"] == result[0]
    assert info["firmware"] == result[1]
    assert info["chainLength"] == result[2]
    assert info["openDirection"] == result[3]
    assert info["fault"] == result[4]
    assert info["solarPanel"] == result[5]
    assert info["calibration"] == result[6]
    assert info["calibrated"] == result[7]
    assert info["inMotion"] == result[8]
    assert info["position"] == result[9]
    assert info["timers"] == result[10]


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_closing(reverse_mode):
    """Test passive closing advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = roller_shade.SwitchbotRollerShade(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 100)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 90)
    )

    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is True


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_opening_then_stop(reverse_mode):
    """Test passive stopped after opening advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = roller_shade.SwitchbotRollerShade(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 0)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 10)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, False, 10)
    )

    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is False


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_closing_then_stop(reverse_mode):
    """Test passive stopped after closing advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = roller_shade.SwitchbotRollerShade(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 100)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 90)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, False, 90)
    )

    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is False


@pytest.mark.asyncio
async def test_stop():
    curtain_device = create_device_for_command_testing()
    await curtain_device.stop()
    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is False
    curtain_device._send_multiple_commands.assert_awaited_once_with(
        roller_shade.STOP_KEYS
    )


@pytest.mark.asyncio
async def test_set_position_opening():
    curtain_device = create_device_for_command_testing(reverse_mode=True)
    await curtain_device.set_position(0)
    assert curtain_device.is_opening() is True
    assert curtain_device.is_closing() is False
    curtain_device._send_multiple_commands.assert_awaited_once()


@pytest.mark.asyncio
async def test_set_position_closing():
    curtain_device = create_device_for_command_testing(reverse_mode=True)
    await curtain_device.set_position(100)
    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is True
    curtain_device._send_multiple_commands.assert_awaited_once()


@pytest.mark.asyncio
async def test_set_position_default_mode_performance():
    """`mode=0` (default) must send the same wire bytes as before quiet mode."""
    curtain_device = create_device_for_command_testing()
    await curtain_device.set_position(50)
    curtain_device._send_multiple_commands.assert_awaited_once_with(
        [
            f"{roller_shade.POSITION_KEYS[0]}32",
            f"{roller_shade.POSITION_KEYS[1]}0032",
        ]
    )


@pytest.mark.asyncio
async def test_set_position_quietdrift():
    """`mode=1` flips the mode byte while leaving the position byte alone."""
    curtain_device = create_device_for_command_testing()
    await curtain_device.set_position(50, mode=1)
    curtain_device._send_multiple_commands.assert_awaited_once_with(
        [
            f"{roller_shade.POSITION_KEYS[0]}32",
            f"{roller_shade.POSITION_KEYS[1]}0132",
        ]
    )


@pytest.mark.asyncio
async def test_set_position_quietdrift_reversed():
    """Quiet mode and reverse mode are independent — both apply correctly."""
    curtain_device = create_device_for_command_testing(reverse_mode=True)
    # position=30 with reverse → device position 70 (0x46), mode byte 01
    await curtain_device.set_position(30, mode=1)
    curtain_device._send_multiple_commands.assert_awaited_once_with(
        [
            f"{roller_shade.POSITION_KEYS[0]}46",
            f"{roller_shade.POSITION_KEYS[1]}0146",
        ]
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("invalid_mode", [-1, 2, 255])
async def test_open_rejects_invalid_mode(invalid_mode):
    roller_shade_device = create_device_for_command_testing()
    with pytest.raises(ValueError, match="mode must be 0"):
        await roller_shade_device.open(mode=invalid_mode)


@pytest.mark.asyncio
@pytest.mark.parametrize("invalid_mode", [-1, 2, 255])
async def test_close_rejects_invalid_mode(invalid_mode):
    roller_shade_device = create_device_for_command_testing()
    with pytest.raises(ValueError, match="mode must be 0"):
        await roller_shade_device.close(mode=invalid_mode)


@pytest.mark.asyncio
@pytest.mark.parametrize("invalid_mode", [-1, 2, 255])
async def test_set_position_rejects_invalid_mode(invalid_mode):
    roller_shade_device = create_device_for_command_testing()
    with pytest.raises(ValueError, match="mode must be 0"):
        await roller_shade_device.set_position(50, mode=invalid_mode)


@pytest.mark.asyncio
async def test_open_does_not_set_motion_flag_on_failure():
    """If the open command fails, _is_opening must remain False."""
    roller_shade_device = create_device_for_command_testing()
    roller_shade_device._send_multiple_commands = AsyncMock(return_value=False)
    result = await roller_shade_device.open()
    assert result is False
    assert roller_shade_device.is_opening() is False
    assert roller_shade_device.is_closing() is False


@pytest.mark.asyncio
async def test_close_speed_kwarg_is_deprecated_alias_for_mode():
    """`close(speed=1)` continues to work but emits DeprecationWarning."""
    roller_shade_device = create_device_for_command_testing()
    with pytest.warns(DeprecationWarning, match="speed.*deprecated"):
        await roller_shade_device.close(speed=1)
    roller_shade_device._send_multiple_commands.assert_awaited_once_with(
        [roller_shade.CLOSE_KEYS[0], f"{roller_shade.CLOSE_KEYS[1]}0164"]
    )


@pytest.mark.asyncio
async def test_close_speed_kwarg_validates_mode():
    """A bad value via `speed=` is still rejected by `_validate_mode`."""
    roller_shade_device = create_device_for_command_testing()
    with (
        pytest.warns(DeprecationWarning, match="speed.*deprecated"),
        pytest.raises(ValueError, match="mode must be 0"),
    ):
        await roller_shade_device.close(speed=2)


@pytest.mark.asyncio
async def test_close_rejects_other_unexpected_kwargs():
    """Unknown kwargs (other than `speed`) should still raise TypeError."""
    roller_shade_device = create_device_for_command_testing()
    with pytest.raises(TypeError, match="unexpected keyword arguments"):
        await roller_shade_device.close(turbo=True)


@pytest.mark.asyncio
async def test_close_does_not_set_motion_flag_on_failure():
    """If the close command fails, _is_closing must remain False."""
    roller_shade_device = create_device_for_command_testing()
    roller_shade_device._send_multiple_commands = AsyncMock(return_value=False)
    result = await roller_shade_device.close()
    assert result is False
    assert roller_shade_device.is_opening() is False
    assert roller_shade_device.is_closing() is False


@pytest.mark.asyncio
async def test_stop_does_not_clear_motion_flags_on_failure():
    """If the stop command fails, prior motion flags persist."""
    roller_shade_device = create_device_for_command_testing()
    roller_shade_device._is_opening = True
    roller_shade_device._send_multiple_commands = AsyncMock(return_value=False)
    result = await roller_shade_device.stop()
    assert result is False
    assert roller_shade_device.is_opening() is True


@pytest.mark.asyncio
async def test_set_position_does_not_update_direction_on_failure():
    """If set_position fails, the motion direction must not be touched."""
    roller_shade_device = create_device_for_command_testing(position=50)
    roller_shade_device._send_multiple_commands = AsyncMock(return_value=False)
    result = await roller_shade_device.set_position(80)
    assert result is False
    assert roller_shade_device.is_opening() is False
    assert roller_shade_device.is_closing() is False


def test_get_position():
    curtain_device = create_device_for_command_testing()
    assert curtain_device.get_position() == 50


def test_update_motion_direction_with_no_previous_position():
    curtain_device = create_device_for_command_testing(reverse_mode=True)
    curtain_device._update_motion_direction(True, None, 100)
    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is False


def test_update_motion_direction_with_previous_position():
    curtain_device = create_device_for_command_testing(reverse_mode=True)
    curtain_device._update_motion_direction(True, 50, 100)
    assert curtain_device.is_opening() is True
    assert curtain_device.is_closing() is False



"""Test keypad vision series device parsing and functionality."""

from unittest.mock import AsyncMock, patch

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement
from switchbot.devices.device import SwitchbotEncryptedDevice
from switchbot.devices.keypad_vision import (
    COMMAND_GET_PASSWORD_COUNT,
    SwitchbotKeypadVision,
)

from . import KEYPAD_VISION_INFO, KEYPAD_VISION_PRO_INFO
from .test_adv_parser import AdvTestCase, generate_ble_device


def create_device_for_command_testing(
    adv_info: AdvTestCase,
    init_data: dict | None = None,
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = SwitchbotKeypadVision(
        ble_device, "ff", "ffffffffffffffffffffffffffffffff", model=adv_info.modelName
    )

    device._send_command = AsyncMock()
    device._send_command_sequence = AsyncMock()
    device.update = AsyncMock()
    device.update_from_advertisement(
        make_advertisement_data(ble_device, adv_info, init_data)
    )
    return device


def make_advertisement_data(
    ble_device: BLEDevice, adv_info: AdvTestCase, init_data: dict | None = None
):
    """Set advertisement data with defaults."""
    if init_data is None:
        init_data = {}

    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": adv_info.service_data,
            "data": adv_info.data | init_data,
            "isEncrypted": False,
            "model": adv_info.model,
            "modelFriendlyName": adv_info.modelFriendlyName,
            "modelName": adv_info.modelName,
        }
        | init_data,
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("adv_info"),
    [
        (KEYPAD_VISION_INFO),
        (KEYPAD_VISION_PRO_INFO),
    ],
)
async def test_get_basic_info_none(adv_info: AdvTestCase) -> None:
    """Test getting basic info returns None when no data."""
    device = create_device_for_command_testing(adv_info)
    device._get_basic_info = AsyncMock(return_value=None)

    info = await device.get_basic_info()
    assert info is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("adv_info", "basic_info", "result"),
    [
        (
            KEYPAD_VISION_INFO,
            b"\x01_\x18\x16\x01\x02\x00\n\x01\x02\x03\x05\x02\x00\x01\x00",
            [95, 2.4, 22, 1, True, True, True, 5, True, False],
        ),
        (
            KEYPAD_VISION_PRO_INFO,
            b"\x01_\x0b\x18\x01\x02\x00\n\x01\x02\x03\x05\x02\x00\x03\x00",
            [95, 1.1, 24, 1, True, True, True, 5, True, True],
        ),
    ],
)
async def test_get_basic_info(
    adv_info: AdvTestCase, basic_info: bytes, result: dict
) -> None:
    """Test getting basic info from Keypad Vision devices."""
    device = create_device_for_command_testing(adv_info)
    device._get_basic_info = AsyncMock(return_value=basic_info)

    info = await device.get_basic_info()
    assert info["battery"] == result[0]
    assert info["firmware"] == result[1]
    assert info["hardware"] == result[2]
    assert info["support_fingerprint"] == result[3]
    assert info["lock_button_enabled"] == result[4]
    assert info["tamper_alarm_enabled"] == result[5]
    assert info["backlight_enabled"] == result[6]
    assert info["backlight_level"] == result[7]
    assert info["prompt_tone_enabled"] == result[8]
    assert info["battery_charging"] == result[9]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "adv_info",
    [
        KEYPAD_VISION_INFO,
        KEYPAD_VISION_PRO_INFO,
    ],
)
async def test_add_invalid_password(adv_info: AdvTestCase) -> None:
    """Test adding an invalid password raises ValueError."""
    device = create_device_for_command_testing(adv_info)

    invalid_passwords = ["123", "abcdef", "1234567890123", "12 3456", "passw0rd!"]

    for password in invalid_passwords:
        with pytest.raises(ValueError, match=r"Password must be 6-12 digits."):
            await device.add_password(password)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "adv_info",
    [
        KEYPAD_VISION_INFO,
        KEYPAD_VISION_PRO_INFO,
    ],
)
@pytest.mark.parametrize(
    ("password", "expected_payload"),
    [
        (
            "123456",
            ["570F52020210FF0006010203040506"],
        ),
        (
            "123456789012",
            ["570F52020220FF000C0102030405060708", "570F5202022109000102"],
        ),
    ],
)
async def test_add_password(
    adv_info: AdvTestCase, password: str, expected_payload: list[str]
) -> None:
    """Test adding a valid password sends correct command."""
    device = create_device_for_command_testing(adv_info)

    await device.add_password(password)

    device._send_command_sequence.assert_awaited_once_with(expected_payload)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "adv_info",
    [
        KEYPAD_VISION_INFO,
        KEYPAD_VISION_PRO_INFO,
    ],
)
async def test_get_password_count_no_response(adv_info: AdvTestCase) -> None:
    """Test getting password count returns None when no response."""
    device = create_device_for_command_testing(adv_info)

    device._send_command.return_value = None

    result = await device.get_password_count()
    device._send_command.assert_awaited_once_with(COMMAND_GET_PASSWORD_COUNT)

    assert result is None


@pytest.mark.asyncio
async def test_get_password_count_for_keypad_vision_pro() -> None:
    """Test getting password count for Keypad Vision Pro."""
    device = create_device_for_command_testing(KEYPAD_VISION_PRO_INFO)

    device._send_command.return_value = bytes(
        [0x01, 0x05, 0x02, 0x03, 0x00, 0x02, 0x01, 0x00]
    )

    result = await device.get_password_count()
    device._send_command.assert_awaited_once_with(COMMAND_GET_PASSWORD_COUNT)

    assert result == {
        "pin": 5,
        "nfc": 2,
        "fingerprint": 3,
        "duress_pin": 0,
        "duress_fingerprint": 2,
        "face": 1,
        "palm_vein": 0,
    }


@pytest.mark.asyncio
async def test_get_password_count_for_keypad_vision() -> None:
    """Test getting password count for Keypad Vision."""
    device = create_device_for_command_testing(KEYPAD_VISION_INFO)

    device._send_command.return_value = bytes([0x01, 0x03, 0x02, 0x01, 0x01, 0x00])

    result = await device.get_password_count()
    device._send_command.assert_awaited_once_with(COMMAND_GET_PASSWORD_COUNT)

    assert result == {
        "pin": 3,
        "nfc": 2,
        "fingerprint": 1,
        "duress_pin": 1,
        "duress_fingerprint": 0,
    }


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "adv_info",
    [
        KEYPAD_VISION_INFO,
        KEYPAD_VISION_PRO_INFO,
    ],
)
@patch.object(SwitchbotEncryptedDevice, "verify_encryption_key", new_callable=AsyncMock)
async def test_verify_encryption_key(
    mock_parent_verify: AsyncMock, adv_info: AdvTestCase
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    key_id = "ff"
    encryption_key = "ffffffffffffffffffffffffffffffff"

    mock_parent_verify.return_value = True

    result = await SwitchbotKeypadVision.verify_encryption_key(
        device=ble_device,
        key_id=key_id,
        encryption_key=encryption_key,
        model=adv_info.modelName,
    )

    mock_parent_verify.assert_awaited_once_with(
        ble_device,
        key_id,
        encryption_key,
        adv_info.modelName,
    )

    assert result is True



from switchbot.adv_parsers.hub3 import calculate_light_intensity


def test_calculate_light_intensity():
    """Test calculating light intensity from Hub 3 light level."""
    # Test valid inputs
    assert calculate_light_intensity(1) == 0
    assert calculate_light_intensity(2) == 50
    assert calculate_light_intensity(5) == 317
    assert calculate_light_intensity(10) == 1023

    # Test invalid inputs
    assert calculate_light_intensity(0) == 0



# Those values have been obtained from the following files in SwitchBot Android app
# That's how you can verify them yourself
# /assets/switchbot_config.json

SWITCHBOT_APP_API_BASE_URL = "api.switchbot.net"
SWITCHBOT_APP_CLIENT_ID = "5nnwmhmsa9xxskm14hd85lm9bm"



from unittest.mock import AsyncMock

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement
from switchbot.adv_parser import SUPPORTED_TYPES
from switchbot.devices import vacuum

from .test_adv_parser import generate_ble_device

common_params = [
    (b".\x00d", ".", 2),
    (b"z\x00\x00", ".", 2),
    (b"3\x00\x00", ".", 2),
    (b"(\x00", "(", 1),
    (b"}\x00", "(", 1),
    (b"\x00\x00M\x00\x10\xfb\xa8", b"\x00\x10\xfb\xa8", 2),
    (b"\x00\x00d\x00\x10\xe0P", b"\x00\x10\xe0P", 2),
]


def create_device_for_command_testing(
    protocol_version: int, rawAdvData: bytes, model: str
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = vacuum.SwitchbotVacuum(ble_device)
    device.update_from_advertisement(
        make_advertisement_data(ble_device, protocol_version, rawAdvData, model)
    )
    device._send_command = AsyncMock()
    device.update = AsyncMock()
    return device


def make_advertisement_data(
    ble_device: BLEDevice, protocol_version: int, rawAdvData: bytes, model: str
):
    """Set advertisement data with defaults."""
    if protocol_version == 1:
        return SwitchBotAdvertisement(
            address="aa:bb:cc:dd:ee:ff",
            data={
                "rawAdvData": rawAdvData,
                "data": {
                    "sequence_number": 2,
                    "dusbin_connected": False,
                    "dustbin_bound": False,
                    "network_connected": True,
                    "battery": 100,
                    "work_status": 0,
                },
                "isEncrypted": False,
                "model": model,
                "modelFriendlyName": SUPPORTED_TYPES[model]["modelFriendlyName"],
                "modelName": SUPPORTED_TYPES[model]["modelName"],
            },
            device=ble_device,
            rssi=-97,
            active=True,
        )
    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": rawAdvData,
            "data": {
                "soc_version": "1.1.083",
                "step": 0,
                "mqtt_connected": True,
                "battery": 100,
                "work_status": 15,
            },
            "isEncrypted": False,
            "model": model,
            "modelFriendlyName": SUPPORTED_TYPES[model]["modelFriendlyName"],
            "modelName": SUPPORTED_TYPES[model]["modelName"],
        },
        device=ble_device,
        rssi=-97,
        active=True,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model"),
    [
        (b".\x00d", "."),
        (b"z\x00\x00", "z"),
        (b"3\x00\x00", "3"),
        (b"\x00\x00M\x00\x10\xfb\xa8", b"\x00\x10\xfb\xa8"),
        (b"\x00\x00d\x00\x10\xe0P", b"\x00\x10\xe0P"),
    ],
)
async def test_status_from_proceess_adv(rawAdvData: bytes, model: str) -> None:
    protocol_version = 2
    device = create_device_for_command_testing(protocol_version, rawAdvData, model)
    assert device.get_soc_version() == "1.1.083"
    assert device.get_last_step() == 0
    assert device.get_mqtt_connnect_status() is True
    assert device.get_battery() == 100
    assert device.get_work_status() == 15


@pytest.mark.asyncio
@pytest.mark.parametrize(("rawAdvData", "model"), [(b"(\x00", "("), (b"}\x00", "}")])
async def test_status_from_proceess_adv_k(rawAdvData: bytes, model: str) -> None:
    protocol_version = 1
    device = create_device_for_command_testing(protocol_version, rawAdvData, model)
    assert device.get_dustbin_bound_status() is False
    assert device.get_dustbin_connnected_status() is False
    assert device.get_network_connected_status() is True
    assert device.get_battery() == 100
    assert device.get_work_status() == 0


@pytest.mark.asyncio
@pytest.mark.parametrize(("rawAdvData", "model", "protocol_version"), common_params)
async def test_clean_up(rawAdvData, model, protocol_version):
    device = create_device_for_command_testing(protocol_version, rawAdvData, model)
    await device.clean_up(protocol_version)
    device._send_command.assert_awaited_once_with(
        vacuum.COMMAND_CLEAN_UP[protocol_version]
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("rawAdvData", "model", "protocol_version"), common_params)
async def test_return_to_dock(rawAdvData, model, protocol_version):
    device = create_device_for_command_testing(protocol_version, rawAdvData, model)
    await device.return_to_dock(protocol_version)
    device._send_command.assert_awaited_once_with(
        vacuum.COMMAND_RETURN_DOCK[protocol_version]
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("rawAdvData", "model", "protocol_version"), common_params)
async def test_get_basic_info_returns_none_when_no_data(
    rawAdvData, model, protocol_version
):
    device = create_device_for_command_testing(protocol_version, rawAdvData, model)
    device._get_basic_info = AsyncMock(return_value=None)

    assert await device.get_basic_info() is None



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement
from switchbot.devices.art_frame import COMMAND_SET_IMAGE, SwitchbotArtFrame

from . import ART_FRAME_INFO
from .test_adv_parser import AdvTestCase, generate_ble_device


def create_device_for_command_testing(
    adv_info: AdvTestCase,
    init_data: dict | None = None,
) -> SwitchbotArtFrame:
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = SwitchbotArtFrame(
        ble_device, "ff", "ffffffffffffffffffffffffffffffff", model=adv_info.modelName
    )
    device.update_from_advertisement(
        make_advertisement_data(ble_device, adv_info, init_data)
    )
    device._send_command = AsyncMock()
    device._check_command_result = MagicMock()
    device.update = AsyncMock()
    return device


def make_advertisement_data(
    ble_device: BLEDevice, adv_info: AdvTestCase, init_data: dict | None = None
) -> SwitchBotAdvertisement:
    """Set advertisement data with defaults."""
    if init_data is None:
        init_data = {}

    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": adv_info.service_data,
            "data": adv_info.data | init_data,
            "isEncrypted": False,
            "model": adv_info.model,
            "modelFriendlyName": adv_info.modelFriendlyName,
            "modelName": adv_info.modelName,
        }
        | init_data,
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
async def test_get_basic_info_none() -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)
    device._get_basic_info = AsyncMock(return_value=None)

    assert await device.get_basic_info() is None

    with pytest.raises(
        RuntimeError, match=r"Failed to retrieve basic info for current image index."
    ):
        await device._get_current_image_index()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("basic_info", "result"),
    [
        (
            b"\x016\x07\x01\x00\x00\x04\x00\xde\x18\xa5\x00\x00\x00\x00\x00\x00",
            [
                False,
                54,
                0.7,
                1,
                0,
                0,
                0,
                0,
                4,
                [0, 222, 24, 165],
            ],
        ),
    ],
)
async def test_get_basic_info_parsing(
    basic_info: str, result: list[bool | int | float | list[int]]
) -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)
    device._get_basic_info = AsyncMock(return_value=basic_info)

    info = await device.get_basic_info()
    assert info["battery_charging"] == result[0]
    assert info["battery"] == result[1]
    assert info["firmware"] == result[2]
    assert info["hardware"] == result[3]
    assert info["display_size"] == result[4]
    assert info["display_mode"] == result[5]
    assert info["last_network_status"] == result[6]
    assert info["current_image_index"] == result[7]
    assert info["total_num_of_images"] == result[8]
    assert info["all_images_index"] == result[9]

    device._update_parsed_data(info)
    assert device.get_all_images_index() == result[9]
    assert device.get_total_images() == result[8]
    assert device.get_current_image_index() == result[7]


@pytest.mark.asyncio
async def test_select_image_with_single_image() -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)

    with (
        patch.object(device, "get_all_images_index", return_value=[1]),
        pytest.raises(RuntimeError, match=r"No images available to select from."),
    ):
        device._select_image_index(1)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("current_index", "all_images_index", "expected_cmd"),
    [
        (100, [1, 100, 150], "150"),
        (150, [1, 100, 150], "1"),
        (1, [1, 100, 150], "100"),
    ],
)
async def test_next_image(
    current_index: int, all_images_index: list[int], expected_cmd: str
) -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)

    with (
        patch.object(device, "get_current_image_index", return_value=current_index),
        patch.object(device, "get_all_images_index", return_value=all_images_index),
    ):
        await device.next_image()
        device._send_command.assert_awaited_with(
            COMMAND_SET_IMAGE.format(f"{int(expected_cmd):02X}")
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("current_index", "all_images_index", "expected_cmd"),
    [
        (100, [1, 100, 150], "1"),
        (150, [1, 100, 150], "100"),
        (1, [1, 100, 150], "150"),
    ],
)
async def test_prev_image(
    current_index: int, all_images_index: list[int], expected_cmd: str
) -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)

    with (
        patch.object(device, "get_current_image_index", return_value=current_index),
        patch.object(device, "get_all_images_index", return_value=all_images_index),
    ):
        await device.prev_image()
        device._send_command.assert_awaited_with(
            COMMAND_SET_IMAGE.format(f"{int(expected_cmd):02X}")
        )


@pytest.mark.asyncio
async def test_set_image_with_invalid_index() -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)

    with (
        patch.object(device, "get_total_images", return_value=3),
        patch.object(device, "get_all_images_index", return_value=[1, 2, 3]),
        pytest.raises(
            ValueError, match=r"Image index 5 is out of range. Total images: 3."
        ),
    ):
        await device.set_image(5)


@pytest.mark.asyncio
async def test_set_image_with_valid_index() -> None:
    device = create_device_for_command_testing(ART_FRAME_INFO)

    with (
        patch.object(device, "get_total_images", return_value=3),
        patch.object(device, "get_all_images_index", return_value=[10, 20, 30]),
    ):
        await device.set_image(1)
        device._send_command.assert_awaited_with(COMMAND_SET_IMAGE.format("14"))


def test_default_model_classvar() -> None:
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = SwitchbotArtFrame(ble_device, "ff", "ffffffffffffffffffffffffffffffff")
    assert device._model == ART_FRAME_INFO.modelName



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import MagicMock, patch

import pytest

import switchbot.discovery as discovery_module
from switchbot.discovery import GetSwitchbotDevices
from switchbot.models import SwitchBotAdvertisement


@dataclass
class _FakeBleakScanner:
    detection_callback: object
    adapter: str

    async def start(self) -> None:
        # detection_callback signature: (device, advertisement_data)
        self.detection_callback(object(), object())
        self.detection_callback(object(), object())

    async def stop(self) -> None:
        return


@pytest.mark.asyncio
async def test_discover_fires_callback_for_each_packet(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: list[SwitchBotAdvertisement] = []

    # Patch parse_advertisement_data to return a different parsed object per invocation.
    parsed: list[SwitchBotAdvertisement] = [
        SwitchBotAdvertisement(
            address="aa:bb:cc:dd:ee:ff",
            data={"model": "c", "modelName": "Curtain", "data": {"position": 10}},
            device=object(),
            rssi=-80,
            active=True,
        ),
        SwitchBotAdvertisement(
            address="aa:bb:cc:dd:ee:ff",
            data={"model": "c", "modelName": "Curtain", "data": {"position": 20}},
            device=object(),
            rssi=-70,
            active=True,
        ),
    ]

    def _fake_parse(_device: object, _advertisement_data: object):
        return parsed.pop(0)

    async def _fake_sleep(_seconds: float) -> None:
        return

    def _fake_bleak_scanner(*, detection_callback, adapter: str):
        return _FakeBleakScanner(detection_callback=detection_callback, adapter=adapter)

    monkeypatch.setattr(discovery_module, "parse_advertisement_data", _fake_parse)
    monkeypatch.setattr(discovery_module.asyncio, "sleep", _fake_sleep)
    monkeypatch.setattr(discovery_module.bleak, "BleakScanner", _fake_bleak_scanner)

    scanner = GetSwitchbotDevices(callback=calls.append)
    result = await scanner.discover(scan_timeout=60)

    assert len(calls) == 2
    assert calls[0].data["data"]["position"] == 10
    assert calls[1].data["data"]["position"] == 20

    # discover() retains backwards compatibility by still returning accumulated data.
    assert result["aa:bb:cc:dd:ee:ff"].data["data"]["position"] == 20


@pytest.mark.asyncio
async def test_callback_exception_does_not_break_discovery(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    adv = SwitchBotAdvertisement(
        address="11:22:33:44:55:66",
        data={"model": "H", "modelName": "Bot", "data": {}},
        device=object(),
        rssi=-50,
        active=True,
    )

    def _fake_parse(_device: object, _advertisement_data: object):
        return adv

    async def _fake_sleep(_seconds: float) -> None:
        return

    def _fake_bleak_scanner(*, detection_callback, adapter: str):
        class _S:
            async def start(self) -> None:
                detection_callback(object(), object())

            async def stop(self) -> None:
                return

        return _S()

    def _boom(_adv: SwitchBotAdvertisement) -> None:
        raise RuntimeError("boom")

    monkeypatch.setattr(discovery_module, "parse_advertisement_data", _fake_parse)
    monkeypatch.setattr(discovery_module.asyncio, "sleep", _fake_sleep)
    monkeypatch.setattr(discovery_module.bleak, "BleakScanner", _fake_bleak_scanner)

    scanner = GetSwitchbotDevices(callback=_boom)
    result = await scanner.discover(scan_timeout=1)

    assert result["11:22:33:44:55:66"] == adv


@pytest.mark.asyncio
async def test_callback_exception_is_logged_and_suppressed(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """
    Test that exceptions raised in the user callback are caught,
    logged, and do not crash the discovery process.
    """
    mock_callback = MagicMock(side_effect=RuntimeError("Boom!"))

    scanner = GetSwitchbotDevices(callback=mock_callback)

    adv = SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={"model": "H", "modelName": "Bot", "data": {}},
        device=MagicMock(),
        rssi=-80,
        active=True,
    )

    with patch("switchbot.discovery.parse_advertisement_data", return_value=adv):
        scanner.detection_callback(MagicMock(), MagicMock())

    mock_callback.assert_called_once()

    assert "Error in discovery callback" in caplog.text
    assert "Boom!" in caplog.text  # Exception message should also be in the log



"""Utility functions for switchbot."""

from functools import lru_cache


@lru_cache(maxsize=512)
def format_mac_upper(mac: str) -> str:
    """Format the mac address string to uppercase with colons."""
    to_test = mac

    if len(to_test) == 17 and to_test.count(":") == 5:
        return to_test.upper()

    if len(to_test) == 17 and to_test.count("-") == 5:
        to_test = to_test.replace("-", "")
    elif len(to_test) == 14 and to_test.count(".") == 2:
        to_test = to_test.replace(".", "")

    if len(to_test) == 12:
        # bare 12-char hex, insert colons
        return ":".join(to_test.upper()[i : i + 2] for i in range(0, 12, 2))

    # Not sure how formatted, return original
    return mac.upper()



from __future__ import annotations

from enum import Enum


class LockStatus(Enum):
    LOCKED = 0
    UNLOCKED = 1
    LOCKING = 2
    UNLOCKING = 3
    LOCKING_STOP = 4  # LOCKING_BLOCKED
    UNLOCKING_STOP = 5  # UNLOCKING_BLOCKED
    NOT_FULLY_LOCKED = 6  # LATCH_LOCKED - Only EU lock type
    HALF_LOCKED = 7  # Only Lock2 EU lock type



from unittest.mock import AsyncMock, MagicMock

import pytest
from bleak.backends.device import BLEDevice

from switchbot import (
    SwitchBotAdvertisement,
    SwitchbotModel,
    SwitchbotOperationError,
)
from switchbot.const.air_purifier import AirPurifierMode
from switchbot.devices import air_purifier

from .test_adv_parser import generate_ble_device

common_params = [
    (b"7\x00\x00\x95-\x00", "7", SwitchbotModel.AIR_PURIFIER_TABLE_US),
    (b"*\x00\x00\x15\x04\x00", "*", SwitchbotModel.AIR_PURIFIER_US),
    (b"+\x00\x00\x15\x04\x00", "+", SwitchbotModel.AIR_PURIFIER_JP),
    (b"8\x00\x00\x95-\x00", "8", SwitchbotModel.AIR_PURIFIER_TABLE_JP),
]


def create_device_for_command_testing(
    rawAdvData: bytes,
    model: str,
    model_name: SwitchbotModel,
    init_data: dict | None = None,
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = air_purifier.SwitchbotAirPurifier(
        ble_device,
        "ff",
        "ffffffffffffffffffffffffffffffff",
        model=model_name,
    )
    device.update = AsyncMock()
    device.update_from_advertisement(
        make_advertisement_data(ble_device, rawAdvData, model, model_name, init_data)
    )
    device._send_command = AsyncMock()
    device._check_command_result = MagicMock()
    return device


def make_advertisement_data(
    ble_device: BLEDevice,
    rawAdvData: bytes,
    model: str,
    model_name: SwitchbotModel,
    init_data: dict | None = None,
):
    """Set advertisement data with defaults."""
    if init_data is None:
        init_data = {}

    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": rawAdvData,
            "data": {
                "isOn": True,
                "mode": "level_3",
                "isAqiValid": False,
                "child_lock": False,
                "speed": 100,
                "aqi_level": "excellent",
                "filter element working time": 405,
                "err_code": 0,
                "sequence_number": 161,
            }
            | init_data,
            "isEncrypted": False,
            "model": model,
            "modelFriendlyName": "Air Purifier",
            "modelName": model_name,
        },
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model", "model_name"),
    common_params,
)
@pytest.mark.parametrize(
    "pm25",
    [150],
)
async def test_status_from_process_adv(rawAdvData, model, model_name, pm25):
    device = create_device_for_command_testing(
        rawAdvData, model, model_name, {"pm25": pm25}
    )
    assert device.get_current_percentage() == 100
    assert device.is_on() is True
    assert device.get_current_aqi_level() == "excellent"
    assert device.get_current_mode() == "level_3"
    assert device.get_current_pm25() == 150


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model", "model_name"),
    common_params,
)
async def test_get_basic_info_returns_none_when_no_data(rawAdvData, model, model_name):
    device = create_device_for_command_testing(rawAdvData, model, model_name)
    device._get_basic_info_by_multi_commands = AsyncMock(return_value=None)

    assert await device.get_basic_info() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model", "model_name"),
    common_params,
)
@pytest.mark.parametrize(
    "mode", ["level_1", "level_2", "level_3", "auto", "pet", "sleep"]
)
async def test_set_preset_mode(rawAdvData, model, model_name, mode):
    device = create_device_for_command_testing(
        rawAdvData, model, model_name, {"mode": mode}
    )
    await device.set_preset_mode(mode)
    assert device.get_current_mode() == mode


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model", "model_name"),
    common_params,
)
async def test_turn_on(rawAdvData, model, model_name):
    device = create_device_for_command_testing(
        rawAdvData, model, model_name, {"isOn": True}
    )
    await device.turn_on()
    assert device.is_on() is True


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model", "model_name"),
    common_params,
)
async def test_turn_off(rawAdvData, model, model_name):
    device = create_device_for_command_testing(
        rawAdvData, model, model_name, {"isOn": False}
    )
    await device.turn_off()
    assert device.is_on() is False


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("rawAdvData", "model", "model_name"),
    common_params,
)
@pytest.mark.parametrize(
    ("response", "expected"),
    [
        (b"\x00", None),
        (b"\x07", None),
        (b"\x01\x02\x03", b"\x01\x02\x03"),
    ],
)
async def test__get_basic_info(rawAdvData, model, model_name, response, expected):
    device = create_device_for_command_testing(rawAdvData, model, model_name)
    device._send_command = AsyncMock(return_value=response)
    result = await device._get_basic_info()
    assert result == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "device_case",
    common_params,
)
@pytest.mark.parametrize(
    "info_case",
    [
        (
            bytearray(
                b"\x01\xa7\xe9\x8c\x08\x00\xb2\x01\x96\x00\x00\x00\x00\x01\x00\x17"
            ),
            bytearray(b"\x01\x01\x11\x22\x33\x44"),
            bytearray(b"\x01\x03"),
            [
                True,
                2,
                "level_2",
                True,
                False,
                "excellent",
                50,
                1,
                2.3,
                0x44,
                True,
            ],
        ),
        (
            bytearray(
                b"\x01\xa8\xec\x8c\x08\x00\xb2\x01\x96\x00\x00\x00\x01\x00\x00\x17"
            ),
            bytearray(b"\x01\x01\xaa\xbb\xcc\x1e"),
            bytearray(b"\x01\x00"),
            [
                True,
                2,
                "pet",
                True,
                False,
                "excellent",
                50,
                256,
                2.3,
                0x1E,
                False,
            ],
        ),
    ],
)
async def test_get_basic_info(device_case, info_case):
    rawAdvData, model, model_name = device_case
    basic_info, led_settings, led_status, result = info_case
    device = create_device_for_command_testing(rawAdvData, model, model_name)
    device._get_basic_info_by_multi_commands = AsyncMock(
        return_value=[basic_info, led_settings, led_status]
    )

    info = await device.get_basic_info()
    assert info["isOn"] == result[0]
    assert info["version_info"] == result[1]
    assert info["mode"] == result[2]
    assert info["isAqiValid"] == result[3]
    assert info["child_lock"] == result[4]
    assert info["aqi_level"] == result[5]
    assert info["speed"] == result[6]
    if model_name not in (
        SwitchbotModel.AIR_PURIFIER_JP,
        SwitchbotModel.AIR_PURIFIER_TABLE_JP,
    ):
        assert info["pm25"] == result[7]
    else:
        assert "pm25" not in info
    assert info["firmware"] == result[8]
    assert info["brightness"] == result[9]
    assert info["light_sensitive"] == result[10]


def test_default_model_classvar():
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = air_purifier.SwitchbotAirPurifier(
        ble_device, "ff", "ffffffffffffffffffffffffffffffff"
    )
    assert device._model == SwitchbotModel.AIR_PURIFIER_US


def test_get_modes():
    assert AirPurifierMode.get_modes() == [
        "level_1",
        "level_2",
        "level_3",
        "auto",
        "sleep",
        "pet",
    ]


@pytest.mark.asyncio
async def test_air_purifier_color_and_led_properties():
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(
        raw_adv,
        model,
        model_name,
    )

    assert device.color_modes == {air_purifier.ColorMode.RGB}
    assert device.color_mode == air_purifier.ColorMode.RGB


@pytest.mark.asyncio
async def test_set_percentage_validation_and_command():
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(
        raw_adv,
        model,
        model_name,
        {"mode": "level_2"},
    )
    device._check_command_result = MagicMock(return_value=True)
    device._send_command = AsyncMock(return_value=b"\x01")

    assert await device.set_percentage(25) is True
    device._send_command.assert_called_with(
        air_purifier.COMMAND_SET_PERCENTAGE.format(percentage=25)
    )

    with pytest.raises(ValueError, match="Percentage must be between 0 and 100"):
        await device.set_percentage(-1)
    with pytest.raises(ValueError, match="Percentage must be between 0 and 100"):
        await device.set_percentage(101)

    invalid_mode_device = create_device_for_command_testing(
        raw_adv,
        model,
        model_name,
        {"mode": "auto"},
    )
    with pytest.raises(ValueError, match="Percentage can only be set in LEVEL modes"):
        await invalid_mode_device.set_percentage(10)


@pytest.mark.asyncio
async def test_set_brightness_validation_and_command():
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(raw_adv, model, model_name)
    device._state = {"r": 1, "g": 2, "b": 3}
    device._check_command_result = MagicMock(return_value=True)
    device._send_command = AsyncMock(return_value=b"\x01")

    assert await device.set_brightness(10) is True
    device._send_command.assert_called_with(
        device._set_brightness_command.format("0102030A")
    )

    with pytest.raises(ValueError, match="Brightness must be between 0 and 100"):
        await device.set_brightness(101)


@pytest.mark.asyncio
async def test_set_rgb_validation_and_command():
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(raw_adv, model, model_name)
    device._check_command_result = MagicMock(return_value=True)
    device._send_command = AsyncMock(return_value=b"\x01")

    assert await device.set_rgb(20, 1, 2, 3) is True
    device._send_command.assert_called_with(device._set_rgb_command.format("01020314"))

    with pytest.raises(ValueError, match="Brightness must be between 0 and 100"):
        await device.set_rgb(101, 1, 2, 3)
    with pytest.raises(ValueError, match="r must be between 0 and 255"):
        await device.set_rgb(10, 256, 2, 3)
    with pytest.raises(ValueError, match="g must be between 0 and 255"):
        await device.set_rgb(10, 1, 256, 3)
    with pytest.raises(ValueError, match="b must be between 0 and 255"):
        await device.set_rgb(10, 1, 2, 256)


@pytest.mark.asyncio
async def test_led_and_light_sensitive_commands():
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(
        raw_adv, model, model_name, {"led_status": True}
    )
    device._check_command_result = MagicMock(return_value=True)
    device._send_command = AsyncMock(return_value=b"\x01")

    assert await device.turn_led_on() is True
    device._send_command.assert_called_with(device._turn_led_on_command)

    assert await device.turn_led_off() is True
    device._send_command.assert_called_with(device._turn_led_off_command)

    assert await device.open_light_sensitive_switch() is True
    device._send_command.assert_called_with(device._open_light_sensitive_switch_command)

    assert await device.close_light_sensitive_switch() is True
    device._send_command.assert_called_with(device._turn_led_on_command)

    device_off = create_device_for_command_testing(
        raw_adv,
        model,
        model_name,
    )
    device_off._check_command_result = MagicMock(return_value=True)
    device_off._send_command = AsyncMock(return_value=b"\x01")
    assert await device_off.close_light_sensitive_switch() is True
    device_off._send_command.assert_called_with(device_off._turn_led_on_command)


@pytest.mark.asyncio
async def test_air_purifier_cache_getters():
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(
        raw_adv,
        model,
        model_name,
        {
            "child_lock": True,
            "wireless_charging": True,
            "light_sensitive": True,
            "speed": 88,
        },
    )

    assert device.get_current_percentage() == 88
    assert device.is_child_lock_on() is True
    assert device.is_wireless_charging_on() is True
    assert device.is_light_sensitive_on() is True


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "operation_case",
    [
        ("open_child_lock", "_open_child_lock_command"),
        ("close_child_lock", "_close_child_lock_command"),
    ],
)
async def test_child_lock_operations(operation_case):
    """Child lock commands should always be forwarded correctly."""
    raw_adv, model, model_name = common_params[0]
    device = create_device_for_command_testing(raw_adv, model, model_name)
    operation_name, command_attr = operation_case
    command = getattr(device, command_attr)

    device._check_function_support = MagicMock()
    device._check_command_result = MagicMock(return_value=True)
    device._send_command = AsyncMock(return_value=b"\x01")

    operation = getattr(device, operation_name)
    assert await operation() is True

    device._check_function_support.assert_called_with(command)
    device._send_command.assert_called_with(command)
    device._check_command_result.assert_called_with(b"\x01", 0, {1})


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("raw_adv", "model", "model_name", "supported"),
    [
        (*common_params[0], True),
        (*common_params[3], True),
        (*common_params[1], False),
        (*common_params[2], False),
    ],
)
async def test_wireless_charging_model_support(raw_adv, model, model_name, supported):
    """Wireless charging operations should only succeed for table variants."""
    device = create_device_for_command_testing(raw_adv, model, model_name)
    if supported:
        device._check_command_result = MagicMock(return_value=True)
        device._send_command = AsyncMock(return_value=b"\x01")

        assert await device.open_wireless_charging() is True
        assert await device.close_wireless_charging() is True

        assert device._send_command.call_args_list == [
            ((device._open_wireless_charging_command,),),
            ((device._close_wireless_charging_command,),),
        ]

    else:
        with pytest.raises(SwitchbotOperationError):
            await device.open_wireless_charging()
        with pytest.raises(SwitchbotOperationError):
            await device.close_wireless_charging()



"""Shared temperature/humidity decoding helpers for T/H sensors."""

from __future__ import annotations

from typing import Any

from ..helpers import celsius_to_fahrenheit


def decode_temp_humidity(temp_data: bytes, battery: int | None) -> dict[str, Any]:
    """
    Decode temperature/humidity/fahrenheit-flag from a 3-byte payload.

    Layout (bytes after company ID, for SwitchBot T/H sensors):
        byte 0: bits[3:0] = temperature decimal (0.1 °C units)
        byte 1: bit[7] = temperature sign (1 = positive), bits[6:0] = integer °C
        byte 2: bit[7] = fahrenheit-display flag, bits[6:0] = humidity %
    """
    _temp_sign = 1 if temp_data[1] & 0b10000000 else -1
    _temp_c = _temp_sign * (
        (temp_data[1] & 0b01111111) + ((temp_data[0] & 0b00001111) / 10)
    )
    _temp_f = celsius_to_fahrenheit(_temp_c)
    _temp_f = (_temp_f * 10) / 10
    humidity = temp_data[2] & 0b01111111

    if _temp_c == 0 and humidity == 0 and battery == 0:
        return {}

    return {
        "temp": {"c": _temp_c, "f": _temp_f},
        "temperature": _temp_c,
        "fahrenheit": bool(temp_data[2] & 0b10000000),
        "humidity": humidity,
        "battery": battery,
    }



from unittest.mock import AsyncMock

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement, SwitchbotModel
from switchbot.devices import blind_tilt
from switchbot.devices.base_cover import COVER_EXT_SUM_KEY

from .test_adv_parser import generate_ble_device


def create_device_for_command_testing(
    position=50, calibration=True, reverse_mode=False
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = blind_tilt.SwitchbotBlindTilt(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, position, calibration)
    )
    curtain_device._send_multiple_commands = AsyncMock()
    curtain_device.update = AsyncMock()
    return curtain_device


def make_advertisement_data(
    ble_device: BLEDevice, in_motion: bool, position: int, calibration: bool = True
):
    """Set advertisement data with defaults."""
    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": b"c\xc0X\x00\x11\x04",
            "data": {
                "calibration": calibration,
                "battery": 88,
                "inMotion": in_motion,
                "tilt": position,
                "lightLevel": 1,
                "deviceChain": 1,
            },
            "isEncrypted": False,
            "model": "c",
            "modelFriendlyName": "Curtain",
            "modelName": SwitchbotModel.CURTAIN,
        },
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
async def test_open():
    blind_device = create_device_for_command_testing()
    await blind_device.open()
    blind_device._send_multiple_commands.assert_awaited_once_with(blind_tilt.OPEN_KEYS)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("position", "keys"),
    [(5, blind_tilt.CLOSE_DOWN_KEYS), (55, blind_tilt.CLOSE_UP_KEYS)],
)
async def test_close(position, keys):
    blind_device = create_device_for_command_testing(position=position)
    await blind_device.close()
    blind_device._send_multiple_commands.assert_awaited_once_with(keys)


@pytest.mark.asyncio
async def test_get_basic_info_returns_none_when_no_data():
    blind_device = create_device_for_command_testing()
    blind_device._get_basic_info = AsyncMock(return_value=None)

    assert await blind_device.get_basic_info() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("reverse_mode", "data", "result"),
    [
        (
            False,
            bytes([0, 1, 10, 2, 255, 255, 50, 4]),
            [1, 1, 1, 1, 1, True, False, False, True, 50, 4],
        ),
        (
            False,
            bytes([0, 1, 10, 2, 0, 0, 50, 4]),
            [1, 1, 0, 0, 0, False, False, False, False, 50, 4],
        ),
        (
            False,
            bytes([0, 1, 10, 2, 0, 1, 50, 4]),
            [1, 1, 0, 0, 1, False, True, False, True, 50, 4],
        ),
        (
            True,
            bytes([0, 1, 10, 2, 255, 255, 50, 4]),
            [1, 1, 1, 1, 1, True, False, True, False, 50, 4],
        ),
        (
            True,
            bytes([0, 1, 10, 2, 0, 0, 50, 4]),
            [1, 1, 0, 0, 0, False, False, False, False, 50, 4],
        ),
        (
            True,
            bytes([0, 1, 10, 2, 0, 1, 50, 4]),
            [1, 1, 0, 0, 1, False, True, False, True, 50, 4],
        ),
    ],
)
async def test_get_basic_info(reverse_mode, data, result):
    blind_device = create_device_for_command_testing(reverse_mode=reverse_mode)
    blind_device._get_basic_info = AsyncMock(return_value=data)

    info = await blind_device.get_basic_info()
    assert info["battery"] == result[0]
    assert info["firmware"] == result[1]
    assert info["light"] == result[2]
    assert info["fault"] == result[2]
    assert info["solarPanel"] == result[3]
    assert info["calibration"] == result[3]
    assert info["calibrated"] == result[3]
    assert info["inMotion"] == result[4]
    assert info["motionDirection"]["opening"] == result[5]
    assert info["motionDirection"]["closing"] == result[6]
    assert info["motionDirection"]["up"] == result[7]
    assert info["motionDirection"]["down"] == result[8]
    assert info["tilt"] == result[9]
    assert info["timers"] == result[10]


@pytest.mark.asyncio
async def test_get_extended_info_summary_sends_command():
    blind_device = create_device_for_command_testing()
    blind_device._send_command = AsyncMock()
    await blind_device.get_extended_info_summary()
    blind_device._send_command.assert_awaited_once_with(key=COVER_EXT_SUM_KEY)


@pytest.mark.asyncio
@pytest.mark.parametrize("data_value", [(None), (b"\x07"), (b"\x00")])
async def test_get_extended_info_summary_returns_none_when_bad_data(data_value):
    blind_device = create_device_for_command_testing()
    blind_device._send_command = AsyncMock(return_value=data_value)
    assert await blind_device.get_extended_info_summary() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("data", "result"), [(bytes([0, 0]), False), (bytes([0, 255]), True)]
)
async def test_get_extended_info_summary(data, result):
    blind_device = create_device_for_command_testing()
    blind_device._send_command = AsyncMock(return_value=data)
    ext_result = await blind_device.get_extended_info_summary()
    assert ext_result["device0"]["light"] == result


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_opening(reverse_mode):
    """Test passive opening advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = blind_tilt.SwitchbotBlindTilt(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 0)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 10)
    )

    assert curtain_device.is_opening() is True
    assert curtain_device.is_closing() is False


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_closing(reverse_mode):
    """Test passive closing advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = blind_tilt.SwitchbotBlindTilt(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 100)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 90)
    )

    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is True


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_opening_then_stop(reverse_mode):
    """Test passive stopped after opening advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = blind_tilt.SwitchbotBlindTilt(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 0)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 10)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, False, 10)
    )

    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is False


@pytest.mark.parametrize("reverse_mode", [(True), (False)])
def test_device_passive_closing_then_stop(reverse_mode):
    """Test passive stopped after closing advertisement."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    curtain_device = blind_tilt.SwitchbotBlindTilt(
        ble_device, reverse_mode=reverse_mode
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 100)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, True, 90)
    )
    curtain_device.update_from_advertisement(
        make_advertisement_data(ble_device, False, 90)
    )

    assert curtain_device.is_opening() is False
    assert curtain_device.is_closing() is False



"""Library to handle connection with Switchbot."""

from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Callable
from functools import lru_cache
from typing import Any, TypedDict

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from .adv_parsers.air_purifier import process_air_purifier
from .adv_parsers.art_frame import process_art_frame
from .adv_parsers.blind_tilt import process_woblindtilt
from .adv_parsers.bot import process_wohand
from .adv_parsers.bulb import process_color_bulb
from .adv_parsers.ceiling_light import process_woceiling
from .adv_parsers.climate_panel import process_climate_panel
from .adv_parsers.contact import process_wocontact
from .adv_parsers.curtain import process_wocurtain
from .adv_parsers.fan import process_fan, process_standing_fan
from .adv_parsers.hub2 import process_wohub2
from .adv_parsers.hub3 import process_hub3
from .adv_parsers.hubmini_matter import process_hubmini_matter
from .adv_parsers.humidifier import process_evaporative_humidifier, process_wohumidifier
from .adv_parsers.keypad import process_wokeypad
from .adv_parsers.keypad_vision import process_keypad_vision, process_keypad_vision_pro
from .adv_parsers.leak import process_leak
from .adv_parsers.light_strip import (
    process_candle_warmer_lamp,
    process_light,
    process_rgbic_light,
    process_rgbicww_ceiling_light,
    process_wostrip,
)
from .adv_parsers.lock import (
    process_lock2,
    process_locklite,
    process_wolock,
    process_wolock_pro,
)
from .adv_parsers.meter import process_wosensorth, process_wosensorth_c
from .adv_parsers.motion import process_wopresence
from .adv_parsers.plug import process_woplugmini
from .adv_parsers.presence_sensor import process_presence_sensor
from .adv_parsers.relay_switch import (
    process_garage_door_opener,
    process_relay_switch_1pm,
    process_relay_switch_2pm,
    process_relay_switch_common_data,
)
from .adv_parsers.remote import process_woremote
from .adv_parsers.roller_shade import process_worollershade
from .adv_parsers.smart_thermostat_radiator import process_smart_thermostat_radiator
from .adv_parsers.vacuum import process_vacuum, process_vacuum_k
from .adv_parsers.weather_station import process_weather_station
from .const import SwitchbotModel
from .models import SwitchBotAdvertisement
from .utils import format_mac_upper

_LOGGER = logging.getLogger(__name__)

SERVICE_DATA_ORDER = (
    "0000fd3d-0000-1000-8000-00805f9b34fb",
    "00000d00-0000-1000-8000-00805f9b34fb",
)
MFR_DATA_ORDER = (2409, 741, 89)

_MODEL_TO_MAC_CACHE: dict[str, SwitchbotModel] = {}


class SwitchbotSupportedType(TypedDict):
    """Supported type of Switchbot."""

    modelName: SwitchbotModel
    modelFriendlyName: str
    func: Callable[[bytes | None, bytes | None], dict[str, bool | int | str | None]]
    manufacturer_id: int | None
    manufacturer_data_length: int | None


SUPPORTED_TYPES: dict[str | bytes, SwitchbotSupportedType] = {
    "d": {
        "modelName": SwitchbotModel.CONTACT_SENSOR,
        "modelFriendlyName": "Contact Sensor",
        "func": process_wocontact,
        "manufacturer_id": 2409,
    },
    "D": {
        "modelName": SwitchbotModel.CONTACT_SENSOR,
        "modelFriendlyName": "Contact Sensor",
        "func": process_wocontact,
        "manufacturer_id": 2409,
    },
    "H": {
        "modelName": SwitchbotModel.BOT,
        "modelFriendlyName": "Bot",
        "func": process_wohand,
        "manufacturer_id": 89,
    },
    "s": {
        "modelName": SwitchbotModel.MOTION_SENSOR,
        "modelFriendlyName": "Motion Sensor",
        "func": process_wopresence,
        "manufacturer_id": 2409,
    },
    "S": {
        "modelName": SwitchbotModel.MOTION_SENSOR,
        "modelFriendlyName": "Motion Sensor",
        "func": process_wopresence,
        "manufacturer_id": 2409,
    },
    "r": {
        "modelName": SwitchbotModel.LIGHT_STRIP,
        "modelFriendlyName": "Light Strip",
        "func": process_wostrip,
        "manufacturer_id": 2409,
    },
    "R": {
        "modelName": SwitchbotModel.LIGHT_STRIP,
        "modelFriendlyName": "Light Strip",
        "func": process_wostrip,
        "manufacturer_id": 2409,
    },
    "{": {
        "modelName": SwitchbotModel.CURTAIN,
        "modelFriendlyName": "Curtain 3",
        "func": process_wocurtain,
        "manufacturer_id": 2409,
    },
    "[": {
        "modelName": SwitchbotModel.CURTAIN,
        "modelFriendlyName": "Curtain 3",
        "func": process_wocurtain,
        "manufacturer_id": 2409,
    },
    "c": {
        "modelName": SwitchbotModel.CURTAIN,
        "modelFriendlyName": "Curtain",
        "func": process_wocurtain,
        "manufacturer_id": 2409,
    },
    "C": {
        "modelName": SwitchbotModel.CURTAIN,
        "modelFriendlyName": "Curtain",
        "func": process_wocurtain,
        "manufacturer_id": 2409,
    },
    "w": {
        "modelName": SwitchbotModel.IO_METER,
        "modelFriendlyName": "Indoor/Outdoor Meter",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "W": {
        "modelName": SwitchbotModel.IO_METER,
        "modelFriendlyName": "Indoor/Outdoor Meter",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "i": {
        "modelName": SwitchbotModel.METER,
        "modelFriendlyName": "Meter Plus",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "I": {
        "modelName": SwitchbotModel.METER,
        "modelFriendlyName": "Meter Plus",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "T": {
        "modelName": SwitchbotModel.METER,
        "modelFriendlyName": "Meter",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "t": {
        "modelName": SwitchbotModel.METER,
        "modelFriendlyName": "Meter",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "4": {
        "modelName": SwitchbotModel.METER_PRO,
        "modelFriendlyName": "Meter Pro",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    b"\x14": {
        "modelName": SwitchbotModel.METER_PRO,
        "modelFriendlyName": "Meter Pro",
        "func": process_wosensorth,
        "manufacturer_id": 2409,
    },
    "5": {
        "modelName": SwitchbotModel.METER_PRO_C,
        "modelFriendlyName": "Meter Pro CO2",
        "func": process_wosensorth_c,
        "manufacturer_id": 2409,
    },
    b"\x15": {
        "modelName": SwitchbotModel.METER_PRO_C,
        "modelFriendlyName": "Meter Pro CO2",
        "func": process_wosensorth_c,
        "manufacturer_id": 2409,
    },
    "v": {
        "modelName": SwitchbotModel.HUB2,
        "modelFriendlyName": "Hub 2",
        "func": process_wohub2,
        "manufacturer_id": 2409,
    },
    "V": {
        "modelName": SwitchbotModel.HUB2,
        "modelFriendlyName": "Hub 2",
        "func": process_wohub2,
        "manufacturer_id": 2409,
    },
    "g": {
        "modelName": SwitchbotModel.PLUG_MINI,
        "modelFriendlyName": "Plug Mini",
        "func": process_woplugmini,
        "manufacturer_id": 2409,
    },
    "G": {
        "modelName": SwitchbotModel.PLUG_MINI,
        "modelFriendlyName": "Plug Mini",
        "func": process_woplugmini,
        "manufacturer_id": 2409,
    },
    "j": {
        "modelName": SwitchbotModel.PLUG_MINI,
        "modelFriendlyName": "Plug Mini (JP)",
        "func": process_woplugmini,
        "manufacturer_id": 2409,
    },
    "J": {
        "modelName": SwitchbotModel.PLUG_MINI,
        "modelFriendlyName": "Plug Mini (JP)",
        "func": process_woplugmini,
        "manufacturer_id": 2409,
    },
    "u": {
        "modelName": SwitchbotModel.COLOR_BULB,
        "modelFriendlyName": "Color Bulb",
        "func": process_color_bulb,
        "manufacturer_id": 2409,
    },
    "U": {
        "modelName": SwitchbotModel.COLOR_BULB,
        "modelFriendlyName": "Color Bulb",
        "func": process_color_bulb,
        "manufacturer_id": 2409,
    },
    "q": {
        "modelName": SwitchbotModel.CEILING_LIGHT,
        "modelFriendlyName": "Ceiling Light",
        "func": process_woceiling,
        "manufacturer_id": 2409,
    },
    "Q": {
        "modelName": SwitchbotModel.CEILING_LIGHT,
        "modelFriendlyName": "Ceiling Light",
        "func": process_woceiling,
        "manufacturer_id": 2409,
    },
    "n": {
        "modelName": SwitchbotModel.CEILING_LIGHT,
        "modelFriendlyName": "Ceiling Light Pro",
        "func": process_woceiling,
        "manufacturer_id": 2409,
    },
    "N": {
        "modelName": SwitchbotModel.CEILING_LIGHT,
        "modelFriendlyName": "Ceiling Light Pro",
        "func": process_woceiling,
        "manufacturer_id": 2409,
    },
    "e": {
        "modelName": SwitchbotModel.HUMIDIFIER,
        "modelFriendlyName": "Humidifier",
        "func": process_wohumidifier,
        "manufacturer_id": 741,
        "manufacturer_data_length": 6,
    },
    "E": {
        "modelName": SwitchbotModel.HUMIDIFIER,
        "modelFriendlyName": "Humidifier",
        "func": process_wohumidifier,
        "manufacturer_id": 741,
        "manufacturer_data_length": 6,
    },
    "#": {
        "modelName": SwitchbotModel.EVAPORATIVE_HUMIDIFIER,
        "modelFriendlyName": "Evaporative Humidifier",
        "func": process_evaporative_humidifier,
        "manufacturer_id": 2409,
    },
    b"\x03": {
        "modelName": SwitchbotModel.EVAPORATIVE_HUMIDIFIER,
        "modelFriendlyName": "Evaporative Humidifier",
        "func": process_evaporative_humidifier,
        "manufacturer_id": 2409,
    },
    "o": {
        "modelName": SwitchbotModel.LOCK,
        "modelFriendlyName": "Lock",
        "func": process_wolock,
        "manufacturer_id": 2409,
    },
    "O": {
        "modelName": SwitchbotModel.LOCK,
        "modelFriendlyName": "Lock",
        "func": process_wolock,
        "manufacturer_id": 2409,
    },
    "$": {
        "modelName": SwitchbotModel.LOCK_PRO,
        "modelFriendlyName": "Lock Pro",
        "func": process_wolock_pro,
        "manufacturer_id": 2409,
    },
    b"\x04": {
        "modelName": SwitchbotModel.LOCK_PRO,
        "modelFriendlyName": "Lock Pro",
        "func": process_wolock_pro,
        "manufacturer_id": 2409,
    },
    "x": {
        "modelName": SwitchbotModel.BLIND_TILT,
        "modelFriendlyName": "Blind Tilt",
        "func": process_woblindtilt,
        "manufacturer_id": 2409,
    },
    "X": {
        "modelName": SwitchbotModel.BLIND_TILT,
        "modelFriendlyName": "Blind Tilt",
        "func": process_woblindtilt,
        "manufacturer_id": 2409,
    },
    "&": {
        "modelName": SwitchbotModel.LEAK,
        "modelFriendlyName": "Leak Detector",
        "func": process_leak,
        "manufacturer_id": 2409,
    },
    b"\x06": {
        "modelName": SwitchbotModel.LEAK,
        "modelFriendlyName": "Leak Detector",
        "func": process_leak,
        "manufacturer_id": 2409,
    },
    "y": {
        "modelName": SwitchbotModel.KEYPAD,
        "modelFriendlyName": "Keypad",
        "func": process_wokeypad,
        "manufacturer_id": 2409,
    },
    "Y": {
        "modelName": SwitchbotModel.KEYPAD,
        "modelFriendlyName": "Keypad",
        "func": process_wokeypad,
        "manufacturer_id": 2409,
    },
    "<": {
        "modelName": SwitchbotModel.RELAY_SWITCH_1PM,
        "modelFriendlyName": "Relay Switch 1PM",
        "func": process_relay_switch_1pm,
        "manufacturer_id": 2409,
    },
    b"\x1c": {
        "modelName": SwitchbotModel.RELAY_SWITCH_1PM,
        "modelFriendlyName": "Relay Switch 1PM",
        "func": process_relay_switch_1pm,
        "manufacturer_id": 2409,
    },
    ";": {
        "modelName": SwitchbotModel.RELAY_SWITCH_1,
        "modelFriendlyName": "Relay Switch 1",
        "func": process_relay_switch_common_data,
        "manufacturer_id": 2409,
    },
    b"\x1b": {
        "modelName": SwitchbotModel.RELAY_SWITCH_1,
        "modelFriendlyName": "Relay Switch 1",
        "func": process_relay_switch_common_data,
        "manufacturer_id": 2409,
    },
    "b": {
        "modelName": SwitchbotModel.REMOTE,
        "modelFriendlyName": "Remote",
        "func": process_woremote,
        "manufacturer_id": 89,
    },
    "B": {
        "modelName": SwitchbotModel.REMOTE,
        "modelFriendlyName": "Remote",
        "func": process_woremote,
        "manufacturer_id": 89,
    },
    ",": {
        "modelName": SwitchbotModel.ROLLER_SHADE,
        "modelFriendlyName": "Roller Shade",
        "func": process_worollershade,
        "manufacturer_id": 2409,
    },
    b"\x0c": {
        "modelName": SwitchbotModel.ROLLER_SHADE,
        "modelFriendlyName": "Roller Shade",
        "func": process_worollershade,
        "manufacturer_id": 2409,
    },
    "%": {
        "modelName": SwitchbotModel.HUBMINI_MATTER,
        "modelFriendlyName": "HubMini Matter",
        "func": process_hubmini_matter,
        "manufacturer_id": 2409,
    },
    b"\x05": {
        "modelName": SwitchbotModel.HUBMINI_MATTER,
        "modelFriendlyName": "HubMini Matter",
        "func": process_hubmini_matter,
        "manufacturer_id": 2409,
    },
    "~": {
        "modelName": SwitchbotModel.CIRCULATOR_FAN,
        "modelFriendlyName": "Circulator Fan",
        "func": process_fan,
        "manufacturer_id": 2409,
    },
    "^": {
        "modelName": SwitchbotModel.CIRCULATOR_FAN,
        "modelFriendlyName": "Circulator Fan",
        "func": process_fan,
        "manufacturer_id": 2409,
    },
    ".": {
        "modelName": SwitchbotModel.K20_VACUUM,
        "modelFriendlyName": "K20 Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    b"\x0f": {
        "modelName": SwitchbotModel.K20_VACUUM,
        "modelFriendlyName": "K20 Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    "z": {
        "modelName": SwitchbotModel.S10_VACUUM,
        "modelFriendlyName": "S10 Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    "Z": {
        "modelName": SwitchbotModel.S10_VACUUM,
        "modelFriendlyName": "S10 Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    "3": {
        "modelName": SwitchbotModel.K10_PRO_COMBO_VACUUM,
        "modelFriendlyName": "K10+ Pro Combo Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    b"\x13": {
        "modelName": SwitchbotModel.K10_PRO_COMBO_VACUUM,
        "modelFriendlyName": "K10+ Pro Combo Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    "}": {
        "modelName": SwitchbotModel.K10_VACUUM,
        "modelFriendlyName": "K10+ Vacuum",
        "func": process_vacuum_k,
        "manufacturer_id": 2409,
    },
    "]": {
        "modelName": SwitchbotModel.K10_VACUUM,
        "modelFriendlyName": "K10+ Vacuum",
        "func": process_vacuum_k,
        "manufacturer_id": 2409,
    },
    "(": {
        "modelName": SwitchbotModel.K10_PRO_VACUUM,
        "modelFriendlyName": "K10+ Pro Vacuum",
        "func": process_vacuum_k,
        "manufacturer_id": 2409,
    },
    b"\x08": {
        "modelName": SwitchbotModel.K10_PRO_VACUUM,
        "modelFriendlyName": "K10+ Pro Vacuum",
        "func": process_vacuum_k,
        "manufacturer_id": 2409,
    },
    "*": {
        "modelName": SwitchbotModel.AIR_PURIFIER_US,
        "modelFriendlyName": "Air Purifier US",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    b"\x0a": {
        "modelName": SwitchbotModel.AIR_PURIFIER_US,
        "modelFriendlyName": "Air Purifier US",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    "+": {
        "modelName": SwitchbotModel.AIR_PURIFIER_JP,
        "modelFriendlyName": "Air Purifier JP",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    b"\x0b": {
        "modelName": SwitchbotModel.AIR_PURIFIER_JP,
        "modelFriendlyName": "Air Purifier JP",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    "7": {
        "modelName": SwitchbotModel.AIR_PURIFIER_TABLE_US,
        "modelFriendlyName": "Air Purifier Table US",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    b"\x17": {
        "modelName": SwitchbotModel.AIR_PURIFIER_TABLE_US,
        "modelFriendlyName": "Air Purifier Table US",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    "8": {
        "modelName": SwitchbotModel.AIR_PURIFIER_TABLE_JP,
        "modelFriendlyName": "Air Purifier Table JP",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    b"\x18": {
        "modelName": SwitchbotModel.AIR_PURIFIER_TABLE_JP,
        "modelFriendlyName": "Air Purifier Table JP",
        "func": process_air_purifier,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xb9\x40": {
        "modelName": SwitchbotModel.HUB3,
        "modelFriendlyName": "Hub3",
        "func": process_hub3,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xb9\x40": {
        "modelName": SwitchbotModel.HUB3,
        "modelFriendlyName": "Hub3",
        "func": process_hub3,
        "manufacturer_id": 2409,
    },
    "-": {
        "modelName": SwitchbotModel.LOCK_LITE,
        "modelFriendlyName": "Lock Lite",
        "func": process_locklite,
        "manufacturer_id": 2409,
    },
    b"\x0d": {
        "modelName": SwitchbotModel.LOCK_LITE,
        "modelFriendlyName": "Lock Lite",
        "func": process_locklite,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xa5\xb8": {
        "modelName": SwitchbotModel.LOCK_ULTRA,
        "modelFriendlyName": "Lock Ultra",
        "func": process_lock2,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xa5\xb8": {
        "modelName": SwitchbotModel.LOCK_ULTRA,
        "modelFriendlyName": "Lock Ultra",
        "func": process_lock2,
        "manufacturer_id": 2409,
    },
    ">": {
        "modelName": SwitchbotModel.GARAGE_DOOR_OPENER,
        "modelFriendlyName": "Garage Door Opener",
        "func": process_garage_door_opener,
        "manufacturer_id": 2409,
    },
    b"\x1e": {
        "modelName": SwitchbotModel.GARAGE_DOOR_OPENER,
        "modelFriendlyName": "Garage Door Opener",
        "func": process_garage_door_opener,
        "manufacturer_id": 2409,
    },
    "=": {
        "modelName": SwitchbotModel.RELAY_SWITCH_2PM,
        "modelFriendlyName": "Relay Switch 2PM",
        "func": process_relay_switch_2pm,
        "manufacturer_id": 2409,
    },
    b"\x1d": {
        "modelName": SwitchbotModel.RELAY_SWITCH_2PM,
        "modelFriendlyName": "Relay Switch 2PM",
        "func": process_relay_switch_2pm,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb0": {
        "modelName": SwitchbotModel.FLOOR_LAMP,
        "modelFriendlyName": "Floor Lamp",
        "func": process_light,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb0": {
        "modelName": SwitchbotModel.FLOOR_LAMP,
        "modelFriendlyName": "Floor Lamp",
        "func": process_light,
        "manufacturer_id": 2409,
    },
    b"\x00\x11\x22\xb8": {
        "modelName": SwitchbotModel.CANDLE_WARMER_LAMP,
        "modelFriendlyName": "Candle Warmer Lamp",
        "func": process_candle_warmer_lamp,
        "manufacturer_id": 2409,
    },
    b"\x01\x11\x22\xb8": {
        "modelName": SwitchbotModel.CANDLE_WARMER_LAMP,
        "modelFriendlyName": "Candle Warmer Lamp",
        "func": process_candle_warmer_lamp,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb1": {
        "modelName": SwitchbotModel.STRIP_LIGHT_3,
        "modelFriendlyName": "Strip Light 3",
        "func": process_light,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb1": {
        "modelName": SwitchbotModel.STRIP_LIGHT_3,
        "modelFriendlyName": "Strip Light 3",
        "func": process_light,
        "manufacturer_id": 2409,
    },
    "?": {
        "modelName": SwitchbotModel.PLUG_MINI_EU,
        "modelFriendlyName": "Plug Mini (EU)",
        "func": process_relay_switch_1pm,
        "manufacturer_id": 2409,
    },
    b"\x1f": {
        "modelName": SwitchbotModel.PLUG_MINI_EU,
        "modelFriendlyName": "Plug Mini (EU)",
        "func": process_relay_switch_1pm,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb3": {
        "modelName": SwitchbotModel.RGBICWW_STRIP_LIGHT,
        "modelFriendlyName": "RGBICWW Strip Light",
        "func": process_rgbic_light,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb3": {
        "modelName": SwitchbotModel.RGBICWW_STRIP_LIGHT,
        "modelFriendlyName": "RGBICWW Strip Light",
        "func": process_rgbic_light,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb4": {
        "modelName": SwitchbotModel.RGBICWW_FLOOR_LAMP,
        "modelFriendlyName": "RGBICWW Floor Lamp",
        "func": process_rgbic_light,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb4": {
        "modelName": SwitchbotModel.RGBICWW_FLOOR_LAMP,
        "modelFriendlyName": "RGBICWW Floor Lamp",
        "func": process_rgbic_light,
        "manufacturer_id": 2409,
    },
    b"\x00\x11\xbb\x10": {
        "modelName": SwitchbotModel.RGBICWW_CEILING_LIGHT,
        "modelFriendlyName": "RGBICWW Ceiling Light",
        "func": process_rgbicww_ceiling_light,
        "manufacturer_id": 2409,
    },
    b"\x01\x11\xbb\x10": {
        "modelName": SwitchbotModel.RGBICWW_CEILING_LIGHT,
        "modelFriendlyName": "RGBICWW Ceiling Light",
        "func": process_rgbicww_ceiling_light,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb7": {
        "modelName": SwitchbotModel.PERMANENT_OUTDOOR_LIGHT,
        "modelFriendlyName": "Permanent Outdoor Light",
        "func": process_rgbic_light,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb7": {
        "modelName": SwitchbotModel.PERMANENT_OUTDOOR_LIGHT,
        "modelFriendlyName": "Permanent Outdoor Light",
        "func": process_rgbic_light,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb5": {
        "modelName": SwitchbotModel.RGBIC_NEON_WIRE_ROPE_LIGHT,
        "modelFriendlyName": "RGBIC Neon Wire Rope Light",
        "func": process_wostrip,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xd0\xb6": {
        "modelName": SwitchbotModel.RGBIC_NEON_ROPE_LIGHT,
        "modelFriendlyName": "RGBIC Neon Rope Light",
        "func": process_wostrip,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb5": {
        "modelName": SwitchbotModel.RGBIC_NEON_WIRE_ROPE_LIGHT,
        "modelFriendlyName": "RGBIC Neon Wire Rope Light",
        "func": process_wostrip,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xd0\xb6": {
        "modelName": SwitchbotModel.RGBIC_NEON_ROPE_LIGHT,
        "modelFriendlyName": "RGBIC Neon Rope Light",
        "func": process_wostrip,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xfb\xa8": {
        "modelName": SwitchbotModel.K11_VACUUM,
        "modelFriendlyName": "K11+ Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xfb\xa8": {
        "modelName": SwitchbotModel.K11_VACUUM,
        "modelFriendlyName": "K11+ Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xf3\xd8": {
        "modelName": SwitchbotModel.CLIMATE_PANEL,
        "modelFriendlyName": "Climate Panel",
        "func": process_climate_panel,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xf3\xd8": {
        "modelName": SwitchbotModel.CLIMATE_PANEL,
        "modelFriendlyName": "Climate Panel",
        "func": process_climate_panel,
        "manufacturer_id": 2409,
    },
    b"\x00\x116@": {
        "modelName": SwitchbotModel.SMART_THERMOSTAT_RADIATOR,
        "modelFriendlyName": "Smart Thermostat Radiator",
        "func": process_smart_thermostat_radiator,
        "manufacturer_id": 2409,
    },
    b"\x01\x116@": {
        "modelName": SwitchbotModel.SMART_THERMOSTAT_RADIATOR,
        "modelFriendlyName": "Smart Thermostat Radiator",
        "func": process_smart_thermostat_radiator,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xe0P": {
        "modelName": SwitchbotModel.S20_VACUUM,
        "modelFriendlyName": "S20 Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xe0P": {
        "modelName": SwitchbotModel.S20_VACUUM,
        "modelFriendlyName": "S20 Vacuum",
        "func": process_vacuum,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xcc\xc8": {
        "modelName": SwitchbotModel.PRESENCE_SENSOR,
        "modelFriendlyName": "Presence Sensor",
        "func": process_presence_sensor,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xcc\xc8": {
        "modelName": SwitchbotModel.PRESENCE_SENSOR,
        "modelFriendlyName": "Presence Sensor",
        "func": process_presence_sensor,
        "manufacturer_id": 2409,
    },
    b"\x00\x11>\x10": {
        "modelName": SwitchbotModel.ART_FRAME,
        "modelFriendlyName": "Art Frame",
        "func": process_art_frame,
        "manufacturer_id": 2409,
    },
    b"\x01\x11>\x10": {
        "modelName": SwitchbotModel.ART_FRAME,
        "modelFriendlyName": "Art Frame",
        "func": process_art_frame,
        "manufacturer_id": 2409,
    },
    b"\x00\x11\x03x": {
        "modelName": SwitchbotModel.KEYPAD_VISION,
        "modelFriendlyName": "Keypad Vision",
        "func": process_keypad_vision,
        "manufacturer_id": 2409,
    },
    b"\x01\x11\x03x": {
        "modelName": SwitchbotModel.KEYPAD_VISION,
        "modelFriendlyName": "Keypad Vision",
        "func": process_keypad_vision,
        "manufacturer_id": 2409,
    },
    b"\x00\x11Q\x98": {
        "modelName": SwitchbotModel.KEYPAD_VISION_PRO,
        "modelFriendlyName": "Keypad Vision Pro",
        "func": process_keypad_vision_pro,
        "manufacturer_id": 2409,
    },
    b"\x01\x11Q\x98": {
        "modelName": SwitchbotModel.KEYPAD_VISION_PRO,
        "modelFriendlyName": "Keypad Vision Pro",
        "func": process_keypad_vision_pro,
        "manufacturer_id": 2409,
    },
    b"\x00\x11\x69\x09": {
        "modelName": SwitchbotModel.LOCK_VISION_PRO,
        "modelFriendlyName": "Lock Vision Pro",
        "func": process_lock2,
        "manufacturer_id": 2409,
    },
    b"\x01\x11\x69\x09": {
        "modelName": SwitchbotModel.LOCK_VISION_PRO,
        "modelFriendlyName": "Lock Vision Pro",
        "func": process_lock2,
        "manufacturer_id": 2409,
    },
    b"\x00\x11\x69\x08": {
        "modelName": SwitchbotModel.LOCK_VISION,
        "modelFriendlyName": "Lock Vision",
        "func": process_locklite,
        "manufacturer_id": 2409,
    },
    b"\x01\x11\x69\x08": {
        "modelName": SwitchbotModel.LOCK_VISION,
        "modelFriendlyName": "Lock Vision",
        "func": process_locklite,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\xff\x90": {
        "modelName": SwitchbotModel.LOCK_PRO_WIFI,
        "modelFriendlyName": "Lock Pro Wifi",
        "func": process_wolock_pro,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\xff\x90": {
        "modelName": SwitchbotModel.LOCK_PRO_WIFI,
        "modelFriendlyName": "Lock Pro Wifi",
        "func": process_wolock_pro,
        "manufacturer_id": 2409,
    },
    b"\x00\x11\x07\x60": {
        "modelName": SwitchbotModel.STANDING_FAN,
        "modelFriendlyName": "Standing Fan",
        "func": process_standing_fan,
        "manufacturer_id": 2409,
    },
    b"\x01\x11\x07\x60": {
        "modelName": SwitchbotModel.STANDING_FAN,
        "modelFriendlyName": "Standing Fan",
        "func": process_standing_fan,
        "manufacturer_id": 2409,
    },
    b"\x00\x10\x53\xb0": {
        "modelName": SwitchbotModel.WEATHER_STATION,
        "modelFriendlyName": "Weather Station",
        "func": process_weather_station,
        "manufacturer_id": 2409,
    },
    b"\x01\x10\x53\xb0": {
        "modelName": SwitchbotModel.WEATHER_STATION,
        "modelFriendlyName": "Weather Station",
        "func": process_weather_station,
        "manufacturer_id": 2409,
    },
}

_SWITCHBOT_MODEL_TO_CHAR: defaultdict[SwitchbotModel, list[str | bytes]] = defaultdict(
    list
)
for model_chr, model_data in SUPPORTED_TYPES.items():
    _SWITCHBOT_MODEL_TO_CHAR[model_data["modelName"]].append(model_chr)

MODELS_BY_MANUFACTURER_DATA: dict[int, list[tuple[str, SwitchbotSupportedType]]] = {
    mfr_id: [] for mfr_id in MFR_DATA_ORDER
}
for model_chr, model in SUPPORTED_TYPES.items():
    if "manufacturer_id" in model:
        mfr_id = model["manufacturer_id"]
        MODELS_BY_MANUFACTURER_DATA[mfr_id].append((model_chr, model))


def parse_advertisement_data(
    device: BLEDevice,
    advertisement_data: AdvertisementData,
    model: SwitchbotModel | None = None,
) -> SwitchBotAdvertisement | None:
    """Parse advertisement data."""
    upper_mac = format_mac_upper(device.address)
    if model is None and upper_mac in _MODEL_TO_MAC_CACHE:
        model = _MODEL_TO_MAC_CACHE[upper_mac]

    service_data = advertisement_data.service_data

    _service_data = None
    for uuid in SERVICE_DATA_ORDER:
        if uuid in service_data:
            _service_data = service_data[uuid]
            break

    _mfr_data = None
    _mfr_id = None
    for mfr_id in MFR_DATA_ORDER:
        if mfr_id in advertisement_data.manufacturer_data:
            _mfr_id = mfr_id
            _mfr_data = advertisement_data.manufacturer_data[mfr_id]
            break

    if _mfr_data is None and _service_data is None:
        return None

    try:
        data = _parse_data(
            _service_data,
            _mfr_data,
            _mfr_id,
            model,
        )
    except Exception:  # pylint: disable=broad-except
        _LOGGER.exception("Failed to parse advertisement data: %s", advertisement_data)
        return None

    if not data:
        return None

    return SwitchBotAdvertisement(
        device.address, data, device, advertisement_data.rssi, bool(_service_data)
    )


def _find_model_from_service_data(_service_data: bytes) -> str | bytes | None:
    """Find model from service data."""
    char_model = chr(_service_data[0] & 0b01111111)
    if char_model in SUPPORTED_TYPES:
        return char_model

    byte_model = bytes([_service_data[0] & 0b01111111])
    if byte_model in SUPPORTED_TYPES:
        return byte_model

    return None


def _find_model_from_switchbot_model(
    _switchbot_model: SwitchbotModel,
) -> str | bytes | None:
    """Find model from switchbot model."""
    if _switchbot_model in _SWITCHBOT_MODEL_TO_CHAR:
        return _SWITCHBOT_MODEL_TO_CHAR[_switchbot_model][0]
    return None


def _find_model_from_manufacturer_data(
    _mfr_id: int, _mfr_data: bytes | None
) -> str | bytes | None:
    """Find model from manufacturer data."""
    if _mfr_id not in MODELS_BY_MANUFACTURER_DATA or _mfr_data is None:
        return None

    for model_chr, model_data in MODELS_BY_MANUFACTURER_DATA[_mfr_id]:
        expected_length = model_data.get("manufacturer_data_length")
        if expected_length is not None and expected_length == len(_mfr_data):
            return model_chr
    return None


def _find_model_from_service_data_suffix(_service_data: bytes) -> str | bytes | None:
    """Find model from service data suffix."""
    if len(_service_data) <= 5:
        return None

    for s in (_service_data[-4:], _service_data[-5:-1]):
        if s in SUPPORTED_TYPES:
            return s
    return None


def build_advertisement_data(
    _model: str | bytes, _service_data: bytes | None, _mfr_data: bytes | None
) -> dict[str, Any]:
    """Build advertisement data dictionary."""
    _isEncrypted = bool(_service_data[0] & 0b10000000) if _service_data else False
    data = {
        "rawAdvData": _service_data,
        "data": {},
        "model": _model,
        "isEncrypted": _isEncrypted,
    }

    type_data = SUPPORTED_TYPES.get(_model)
    if type_data:
        model_data = type_data["func"](_service_data, _mfr_data)
        if model_data:
            data.update(
                {
                    "modelFriendlyName": type_data["modelFriendlyName"],
                    "modelName": type_data["modelName"],
                    "data": model_data,
                }
            )

    return data


@lru_cache(maxsize=128)
def _parse_data(
    _service_data: bytes | None,
    _mfr_data: bytes | None,
    _mfr_id: int | None = None,
    _switchbot_model: SwitchbotModel | None = None,
) -> dict[str, Any] | None:
    """Parse advertisement data."""
    _model = None

    if _service_data:
        _model = _find_model_from_service_data(_service_data)

    if not _model and _switchbot_model:
        _model = _find_model_from_switchbot_model(_switchbot_model)

    if not _model and _mfr_id:
        _model = _find_model_from_manufacturer_data(_mfr_id, _mfr_data)

    if not _model and _service_data:
        _model = _find_model_from_service_data_suffix(_service_data)

    if not _model:
        return None

    return build_advertisement_data(_model, _service_data, _mfr_data)


def populate_model_to_mac_cache(mac: str, model: SwitchbotModel) -> None:
    """Populate the model to MAC address cache."""
    _MODEL_TO_MAC_CACHE[mac] = model



from enum import Enum


class ColorMode(Enum):
    OFF = 0
    COLOR_TEMP = 1
    RGB = 2
    EFFECT = 3
    BRIGHTNESS = 4


class StripLightColorMode(Enum):
    RGB = 2
    SCENE = 3
    MUSIC = 4
    CONTROLLER = 5
    COLOR_TEMP = 6
    UNKNOWN = 10


class BulbColorMode(Enum):
    COLOR_TEMP = 1
    RGB = 2
    DYNAMIC = 3
    UNKNOWN = 10


class CeilingLightColorMode(Enum):
    COLOR_TEMP = 0
    NIGHT = 1
    MUSIC = 4
    UNKNOWN = 10


class RGBICStripLightColorMode(Enum):
    SEGMENTED = 1
    RGB = 2
    SCENE = 3
    MUSIC = 4
    CONTROLLER = 5
    COLOR_TEMP = 6
    EFFECT = 7
    UNKNOWN = 10


class RGBICWWCeilingLightColorMode(Enum):
    SEGMENTED = 1
    COLOR = 2
    SCENE = 3
    MUSIC = 4
    CONTROLLER = 5
    WARMWHITE = 6
    EFFECT = 7
    UNKNOWN = 10


DEFAULT_COLOR_TEMP = 4001



"""Smart Thermostat Radiator"""

import logging

from ..const.climate import SmartThermostatRadiatorMode

_LOGGER = logging.getLogger(__name__)


def process_smart_thermostat_radiator(
    data: bytes | None, mfr_data: bytes | None
) -> dict[str, bool | int | str]:
    """Process Smart Thermostat Radiator data."""
    if mfr_data is None or len(mfr_data) < 13:
        return {}

    _seq_num = mfr_data[6]
    _isOn = bool(mfr_data[7] & 0b10000000)
    _battery = mfr_data[7] & 0b01111111

    temp_data = mfr_data[8:11]
    target_decimal = (temp_data[0] >> 4) & 0x0F
    local_decimal = temp_data[0] & 0x0F

    local_sign = 1 if (temp_data[1] & 0x80) else -1
    local_int = temp_data[1] & 0x7F
    local_temp = local_sign * (local_int + (local_decimal / 10))

    target_sign = 1 if (temp_data[2] & 0x80) else -1
    target_int = temp_data[2] & 0x7F
    target_temp = target_sign * (target_int + (target_decimal / 10))

    last_mode = SmartThermostatRadiatorMode.get_mode_name((mfr_data[11] >> 4) & 0x0F)
    mode = SmartThermostatRadiatorMode.get_mode_name(mfr_data[11] & 0x07)

    need_update_temp = bool((mfr_data[12] >> 5) & 0x01)
    restarted = bool((mfr_data[12] >> 4) & 0x01)
    fault_code = (mfr_data[12] >> 1) & 0x07
    door_open = bool(mfr_data[12] & 0x01)

    result = {
        "sequence_number": _seq_num,
        "isOn": _isOn,
        "battery": _battery,
        "temperature": local_temp,
        "target_temperature": target_temp,
        "mode": mode,
        "last_mode": last_mode,
        "need_update_temp": need_update_temp,
        "restarted": restarted,
        "fault_code": fault_code,
        "door_open": door_open,
    }

    _LOGGER.debug(
        "Smart Thermostat Radiator mfr data: %s, result: %s", mfr_data.hex(), result
    )
    return result



"""Tests for SwitchbotEncryptedDevice base class."""

from __future__ import annotations

import asyncio
from typing import Any
from unittest.mock import AsyncMock, patch

import pytest
from bleak.exc import BleakDBusError

from switchbot import SwitchbotModel
from switchbot.devices.device import AESMode, SwitchbotEncryptedDevice

from .test_adv_parser import generate_ble_device


class MockEncryptedDevice(SwitchbotEncryptedDevice):
    """Mock encrypted device for testing."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.update_count: int = 0

    async def update(self, interface: int | None = None) -> None:
        self.update_count += 1


def create_encrypted_device(
    model: SwitchbotModel = SwitchbotModel.LOCK,
) -> MockEncryptedDevice:
    """Create an encrypted device for testing."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "Test Device")
    return MockEncryptedDevice(
        ble_device, "01", "0123456789abcdef0123456789abcdef", model=model
    )


@pytest.mark.asyncio
async def test_encrypted_device_init() -> None:
    """Test encrypted device initialization."""
    device = create_encrypted_device()
    assert device._key_id == "01"
    assert device._encryption_key == bytearray.fromhex(
        "0123456789abcdef0123456789abcdef"
    )
    assert device._iv is None
    assert device._cipher is None


@pytest.mark.asyncio
async def test_encrypted_device_init_validation() -> None:
    """Test encrypted device initialization with invalid parameters."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "Test Device")

    # Test empty key_id
    with pytest.raises(ValueError, match="key_id is missing"):
        MockEncryptedDevice(
            ble_device,
            "",
            "0123456789abcdef0123456789abcdef",
            model=SwitchbotModel.LOCK,
        )

    # Test invalid key_id length
    with pytest.raises(ValueError, match="key_id is invalid"):
        MockEncryptedDevice(
            ble_device,
            "1",
            "0123456789abcdef0123456789abcdef",
            model=SwitchbotModel.LOCK,
        )

    # Test empty encryption_key
    with pytest.raises(ValueError, match="encryption_key is missing"):
        MockEncryptedDevice(ble_device, "01", "", model=SwitchbotModel.LOCK)

    # Test invalid encryption_key length
    with pytest.raises(ValueError, match="encryption_key is invalid"):
        MockEncryptedDevice(
            ble_device, "01", "0123456789abcdef", model=SwitchbotModel.LOCK
        )


@pytest.mark.asyncio
async def test_send_command_unencrypted() -> None:
    """Test sending unencrypted command."""
    device = create_encrypted_device()

    with patch.object(device, "_send_command_locked_with_retry") as mock_send:
        mock_send.return_value = b"\x01\x00\x00\x00"

        result = await device._send_command("570200", encrypt=False)

        assert result == b"\x01\x00\x00\x00"
        mock_send.assert_called_once()
        # Verify the key was padded with zeros for unencrypted command
        call_args = mock_send.call_args[0]
        assert call_args[0] == "570000000200"  # Original key with zeros inserted


@pytest.mark.asyncio
async def test_send_command_encrypted_success() -> None:
    """Test successful encrypted command."""
    device = create_encrypted_device()

    # Mock the connection and command execution
    with (
        patch.object(device, "_send_command_locked_with_retry") as mock_send,
        patch.object(device, "_decrypt") as mock_decrypt,
    ):
        mock_decrypt.return_value = b"decrypted_response"

        # First call is for IV initialization, second is for the actual command
        mock_send.side_effect = [
            b"\x01\x00\x00\x00\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0",  # IV response (16 bytes)
            b"\x01\x00\x00\x00encrypted_response",  # Command response
        ]

        result = await device._send_command("570200", encrypt=True)

        assert result is not None
        assert mock_send.call_count == 2
        # Verify IV was initialized
        assert device._iv is not None


@pytest.mark.asyncio
async def test_send_command_iv_already_initialized() -> None:
    """Test sending encrypted command when IV is already initialized."""
    device = create_encrypted_device()

    # Pre-set the IV
    device._iv = b"\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0"

    with (
        patch.object(device, "_send_command_locked_with_retry") as mock_send,
        patch.object(device, "_encrypt") as mock_encrypt,
        patch.object(device, "_decrypt") as mock_decrypt,
    ):
        mock_encrypt.return_value = (
            "656e637279707465645f64617461",  # "encrypted_data" in hex
            "abcd",
        )
        mock_decrypt.return_value = b"decrypted_response"
        mock_send.return_value = b"\x01\x00\x00\x00encrypted_response"

        result = await device._send_command("570200", encrypt=True)

        assert result == b"\x01decrypted_response"
        # Should only call once since IV is already initialized
        mock_send.assert_called_once()
        mock_encrypt.assert_called_once()
        mock_decrypt.assert_called_once()


@pytest.mark.asyncio
async def test_iv_race_condition_during_disconnect() -> None:
    """Test that commands during disconnect are handled properly."""
    device = create_encrypted_device()

    # Pre-set the IV
    device._iv = b"\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78"

    # Mock the connection
    mock_client = AsyncMock()
    mock_client.is_connected = True
    device._client = mock_client

    async def simulate_disconnect() -> None:
        """Simulate disconnect happening during command execution."""
        await asyncio.sleep(0.01)  # Small delay
        await device._execute_disconnect()

    with (
        patch.object(device, "_send_command_locked_with_retry") as mock_send,
        patch.object(device, "_ensure_connected"),
        patch.object(device, "_encrypt") as mock_encrypt,
        patch.object(device, "_decrypt") as mock_decrypt,
    ):
        mock_encrypt.return_value = ("656e63727970746564", "abcd")
        mock_decrypt.return_value = b"response"
        mock_send.return_value = b"\x01\x00\x00\x00response"

        # Start command and disconnect concurrently
        command_task = asyncio.create_task(device._send_command("570200"))
        disconnect_task = asyncio.create_task(simulate_disconnect())

        # Both should complete without error
        result, _ = await asyncio.gather(
            command_task, disconnect_task, return_exceptions=True
        )

        # Command should have completed successfully
        assert isinstance(result, bytes) or result is None
        # IV should be cleared after disconnect
        assert device._iv is None


@pytest.mark.asyncio
async def test_ensure_encryption_initialized_with_lock_held() -> None:
    """Test that _ensure_encryption_initialized properly handles the operation lock."""
    device = create_encrypted_device()

    # Acquire the operation lock
    async with device._operation_lock:
        with patch.object(device, "_send_command_locked_with_retry") as mock_send:
            mock_send.return_value = b"\x01\x00\x00\x00\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0"

            result = await device._ensure_encryption_initialized()

            assert result is True
            assert (
                device._iv
                == b"\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0"
            )
            assert device._cipher is None  # Should be reset when IV changes


@pytest.mark.asyncio
async def test_ensure_encryption_initialized_sets_gcm_mode() -> None:
    """Test that GCM mode is detected from device response."""
    device = create_encrypted_device()

    gcm_iv = b"\x01" * 12
    response = b"\x01\x00\x01\x00" + gcm_iv + b"\x00\x00\x00\x00"

    async with device._operation_lock:
        with patch.object(device, "_send_command_locked_with_retry") as mock_send:
            mock_send.return_value = response

            result = await device._ensure_encryption_initialized()

            assert result is True
            assert device._encryption_mode == AESMode.GCM
            assert device._iv == gcm_iv


@pytest.mark.asyncio
async def test_ensure_encryption_initialized_invalid_iv_length_gcm() -> None:
    """Test that invalid IV length for GCM mode returns False."""
    device = create_encrypted_device()

    # GCM expects 12 bytes IV, but response has wrong length (only 8 bytes after trimming)
    response = b"\x01\x00\x01\x00" + b"\x01" * 8 + b"\x00\x00\x00\x00"

    async with device._operation_lock:
        with patch.object(device, "_send_command_locked_with_retry") as mock_send:
            mock_send.return_value = response

            result = await device._ensure_encryption_initialized()

            assert result is False
            assert device._iv is None


@pytest.mark.asyncio
async def test_ensure_encryption_initialized_invalid_iv_length_ctr() -> None:
    """Test that invalid IV length for CTR mode returns False."""
    device = create_encrypted_device()

    # CTR expects 16 bytes IV, but response has only 8 bytes
    response = b"\x01\x00\x00\x00" + b"\x01" * 8

    async with device._operation_lock:
        with patch.object(device, "_send_command_locked_with_retry") as mock_send:
            mock_send.return_value = response

            result = await device._ensure_encryption_initialized()

            assert result is False
            assert device._iv is None


@pytest.mark.asyncio
async def test_device_with_gcm_mode() -> None:
    """Test that device initializes correctly in GCM mode and increments GCM IV."""
    device = create_encrypted_device()
    device._encryption_mode = AESMode.GCM
    device._iv = b"\x01" * 12

    with (
        patch.object(device, "_ensure_encryption_initialized") as mock_ensure,
        patch.object(device, "_send_command_locked_with_retry") as mock_send,
        patch.object(device, "_decrypt") as mock_decrypt,
        patch.object(device, "_encrypt") as mock_encrypt,
        patch.object(device, "_increment_gcm_iv") as mock_inc_iv,
    ):
        mock_ensure.return_value = True
        mock_encrypt.return_value = ("10203040", "abcd")
        mock_send.return_value = b"\x01\x00\x00\x00\x10\x20\x30\x40"
        mock_decrypt.return_value = b"\x10\x20\x30\x40"

        await device._send_command("570200")

        mock_inc_iv.assert_called_once()


@pytest.mark.asyncio
async def test_resolve_encryption_mode_invalid() -> None:
    """Test that invalid mode byte raises error."""
    device = create_encrypted_device()

    with pytest.raises(ValueError, match="Unsupported encryption mode"):
        device._resolve_encryption_mode(2)


@pytest.mark.asyncio
async def test_resolve_encryption_mode_missing() -> None:
    """Test that missing mode byte raises error."""
    device = create_encrypted_device()

    with pytest.raises(ValueError, match="Encryption mode byte is missing"):
        device._resolve_encryption_mode(None)


@pytest.mark.asyncio
async def test_resolve_encryption_mode_conflict() -> None:
    """Test that conflicting encryption modes raise error."""
    device = create_encrypted_device()
    device._encryption_mode = AESMode.CTR

    with pytest.raises(
        ValueError,
        match="Conflicting encryption modes detected: CTR vs GCM",
    ):
        device._resolve_encryption_mode(1)


@pytest.mark.asyncio
async def test_increment_gcm_iv() -> None:
    """Test GCM IV increment logic."""
    device = create_encrypted_device()
    device._encryption_mode = AESMode.GCM
    device._iv = b"\x00" * 11 + b"\x01"

    device._increment_gcm_iv()

    assert device._iv == b"\x00" * 11 + b"\x02"
    assert device._cipher is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("initial_iv", "expected_exception", "expected_message"),
    [
        (None, RuntimeError, "Cannot increment GCM IV: IV is None"),
        (
            b"\x00" * 10,
            RuntimeError,
            "Cannot increment GCM IV: IV length is not 12 bytes",
        ),
    ],
)
async def test_increment_gcm_iv_invalid(
    initial_iv, expected_exception, expected_message
) -> None:
    """Test GCM IV increment with invalid IV states."""
    device = create_encrypted_device()
    device._encryption_mode = AESMode.GCM
    device._iv = initial_iv

    with pytest.raises(expected_exception, match=expected_message):
        device._increment_gcm_iv()


@pytest.mark.asyncio
async def test_gcm_encrypt_decrypt_without_finalize() -> None:
    """Test GCM encrypt/decrypt works without finalize in decrypt."""
    device = create_encrypted_device()
    device._encryption_mode = AESMode.GCM
    device._iv = b"\x10" * 12

    ciphertext_hex, _ = device._encrypt("48656c6c6f")
    decrypted = device._decrypt(bytearray.fromhex(ciphertext_hex))

    assert decrypted.hex() == "48656c6c6f"


@pytest.mark.asyncio
async def test_ensure_encryption_initialized_failure() -> None:
    """Test _ensure_encryption_initialized when IV initialization fails."""
    device = create_encrypted_device()

    async with device._operation_lock:
        with patch.object(device, "_send_command_locked_with_retry") as mock_send:
            # Return failure response
            mock_send.return_value = b"\x00"

            result = await device._ensure_encryption_initialized()

            assert result is False
            assert device._iv is None


@pytest.mark.asyncio
async def test_encrypt_decrypt_with_valid_iv() -> None:
    """Test encryption and decryption with valid IV."""
    device = create_encrypted_device()
    device._iv = b"\x00" * 16  # Use zeros for predictable test

    # Test encryption
    ciphertext_hex, header_hex = device._encrypt("48656c6c6f")  # "Hello" in hex
    assert isinstance(ciphertext_hex, str)
    assert isinstance(header_hex, str)
    assert len(ciphertext_hex) > 0

    # Test decryption
    decrypted = device._decrypt(bytearray.fromhex(ciphertext_hex))
    assert decrypted.hex() == "48656c6c6f"


@pytest.mark.asyncio
async def test_encrypt_with_none_iv() -> None:
    """Test that encryption raises error when IV is None."""
    device = create_encrypted_device()
    device._iv = None

    with pytest.raises(RuntimeError, match="Cannot encrypt: IV is None"):
        device._encrypt("48656c6c6f")


@pytest.mark.asyncio
async def test_decrypt_with_none_iv() -> None:
    """Test that decryption raises error when IV is None."""
    device = create_encrypted_device()
    device._iv = None

    with pytest.raises(RuntimeError, match="Cannot decrypt: IV is None"):
        device._decrypt(bytearray.fromhex("48656c6c6f"))


@pytest.mark.asyncio
async def test_get_cipher_with_none_iv() -> None:
    """Test that _get_cipher raises error when IV is None."""
    device = create_encrypted_device()
    device._iv = None

    with pytest.raises(RuntimeError, match="Cannot create cipher: IV is None"):
        device._get_cipher()


@pytest.mark.asyncio
async def test_execute_disconnect_clears_encryption_state() -> None:
    """Test that disconnect properly clears encryption state."""
    device = create_encrypted_device()
    device._iv = b"\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0"
    device._cipher = None  # type: ignore[assignment]
    device._encryption_mode = AESMode.CTR

    # Mock client
    mock_client = AsyncMock()
    device._client = mock_client

    with patch.object(device, "_execute_disconnect_with_lock") as mock_disconnect:
        await device._execute_disconnect()

    assert device._iv is None
    assert device._cipher is None
    assert device._encryption_mode is None
    mock_disconnect.assert_called_once()


@pytest.mark.asyncio
async def test_concurrent_commands_with_same_device() -> None:
    """Test multiple concurrent commands on the same device."""
    device = create_encrypted_device()

    # Pre-initialize IV (16 bytes for AES CTR mode)
    device._iv = b"\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0"

    with (
        patch.object(device, "_send_command_locked_with_retry") as mock_send,
        patch.object(device, "_encrypt") as mock_encrypt,
        patch.object(device, "_decrypt") as mock_decrypt,
    ):
        mock_encrypt.return_value = ("656e63727970746564", "abcd")
        mock_decrypt.return_value = b"response"
        mock_send.return_value = b"\x01\x00\x00\x00data"

        # Send multiple commands concurrently
        tasks = [
            device._send_command("570200"),
            device._send_command("570201"),
            device._send_command("570202"),
        ]

        results = await asyncio.gather(*tasks)

        # All commands should succeed
        assert all(result == b"\x01response" for result in results)
        assert mock_send.call_count == 3


@pytest.mark.asyncio
async def test_command_retry_with_encryption() -> None:
    """Test command retry logic with encrypted commands."""
    device = create_encrypted_device()
    device._retry_count = 2

    # Pre-initialize IV (16 bytes for AES CTR mode)
    device._iv = b"\x12\x34\x56\x78\x9a\xbc\xde\xf0\x12\x34\x56\x78\x9a\xbc\xde\xf0"

    with (
        patch.object(device, "_send_command_locked") as mock_send_locked,
        patch.object(device, "_ensure_connected"),
        patch.object(device, "_encrypt") as mock_encrypt,
        patch.object(device, "_decrypt") as mock_decrypt,
    ):
        mock_encrypt.return_value = ("656e63727970746564", "abcd")
        mock_decrypt.return_value = b"response"

        # First attempt fails, second succeeds
        mock_send_locked.side_effect = [
            BleakDBusError("org.bluez.Error", []),
            b"\x01\x00\x00\x00data",
        ]

        result = await device._send_command("570200")

        assert result == b"\x01response"
        assert mock_send_locked.call_count == 2


@pytest.mark.asyncio
async def test_empty_data_encryption_decryption() -> None:
    """Test encryption/decryption of empty data."""
    device = create_encrypted_device()
    device._iv = b"\x00" * 16

    # Test empty encryption
    encrypted = device._encrypt("")
    assert encrypted == ("", "")

    # Test empty decryption
    decrypted = device._decrypt(bytearray())
    assert decrypted == b""


@pytest.mark.asyncio
async def test_verify_encryption_key_falls_back_to_classvar() -> None:
    """verify_encryption_key resolves `model` from `cls._model` when omitted."""

    class ClassvarModelDevice(MockEncryptedDevice):
        _model = SwitchbotModel.LOCK

    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "Test Device")
    with patch.object(
        ClassvarModelDevice,
        "get_basic_info",
        new_callable=AsyncMock,
        return_value={"ok": True},
    ):
        result = await ClassvarModelDevice.verify_encryption_key(
            ble_device, "01", "0123456789abcdef0123456789abcdef"
        )
    assert result is True


@pytest.mark.asyncio
async def test_verify_encryption_key_without_model_or_classvar_raises() -> None:
    """verify_encryption_key raises when neither `model=` nor `cls._model` is set."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "Test Device")
    with pytest.raises(ValueError, match="model must be provided"):
        await MockEncryptedDevice.verify_encryption_key(
            ble_device, "01", "0123456789abcdef0123456789abcdef"
        )


def test_init_without_model_or_classvar_raises() -> None:
    """__init__ raises when neither `model=` nor `cls._model` is set."""
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "Test Device")
    with pytest.raises(ValueError, match="model must be provided"):
        MockEncryptedDevice(ble_device, "01", "0123456789abcdef0123456789abcdef")


@pytest.mark.asyncio
async def test_decrypt_with_none_iv_during_disconnect() -> None:
    """Test that decryption returns empty bytes when IV is None during expected disconnect."""
    device = create_encrypted_device()

    # Simulate disconnection in progress
    device._expected_disconnect = True
    device._iv = None

    # Should return empty bytes instead of raising
    result = device._decrypt(bytearray(b"encrypted_data"))
    assert result == b""

    # Verify it still raises when not disconnecting
    device._expected_disconnect = False
    with pytest.raises(RuntimeError, match="Cannot decrypt: IV is None"):
        device._decrypt(bytearray(b"encrypted_data"))



"""Device handler for the Art Frame."""

import logging
from typing import Any

from ..const import SwitchbotModel
from .device import (
    SwitchbotEncryptedDevice,
    SwitchbotSequenceDevice,
    update_after_operation,
)

_LOGGER = logging.getLogger(__name__)

COMMAND_SET_IMAGE = "570F7A02{}"


class SwitchbotArtFrame(SwitchbotSequenceDevice, SwitchbotEncryptedDevice):
    """Representation of a Switchbot Art Frame."""

    _model = SwitchbotModel.ART_FRAME
    response_flag: bool = True

    async def get_basic_info(self) -> dict[str, Any] | None:
        """Get device basic settings."""
        if not (_data := await self._get_basic_info()):
            return None
        _LOGGER.debug("basic info data: %s", _data.hex())

        battery_charging = bool(_data[1] & 0x80)
        battery = _data[1] & 0x7F
        firmware = _data[2] / 10.0
        hardware = _data[3]
        display_size = (_data[4] >> 4) & 0x0F
        display_mode = (_data[4] >> 3) & 0x01
        last_network_status = (_data[4] >> 2) & 0x01
        current_image_index = _data[5]
        total_num_of_images = _data[6]
        all_images_index = [_data[x] for x in range(7, 7 + total_num_of_images)]

        basic_info = {
            "battery_charging": battery_charging,
            "battery": battery,
            "firmware": firmware,
            "hardware": hardware,
            "display_size": display_size,
            "display_mode": display_mode,
            "last_network_status": last_network_status,
            "current_image_index": current_image_index,
            "total_num_of_images": total_num_of_images,
            "all_images_index": all_images_index,
        }
        _LOGGER.debug("Art Frame %s basic info: %s", self._device.address, basic_info)
        return basic_info

    def _select_image_index(self, offset: int) -> int:
        """Select the image index based on the current index and offset."""
        current_index = self.get_current_image_index()
        all_images_index = self.get_all_images_index()

        if not all_images_index or len(all_images_index) <= 1:
            raise RuntimeError("No images available to select from.")

        new_position = (all_images_index.index(current_index) + offset) % len(
            all_images_index
        )
        return all_images_index[new_position]

    async def _get_current_image_index(self) -> None:
        """Validate the current image index."""
        if not await self.get_basic_info():
            raise RuntimeError("Failed to retrieve basic info for current image index.")

    @update_after_operation
    async def next_image(self) -> bool:
        """Display the next image."""
        await self._get_current_image_index()
        idx = self._select_image_index(1)
        result = await self._send_command(COMMAND_SET_IMAGE.format(f"{idx:02X}"))
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def prev_image(self) -> bool:
        """Display the previous image."""
        await self._get_current_image_index()
        idx = self._select_image_index(-1)
        result = await self._send_command(COMMAND_SET_IMAGE.format(f"{idx:02X}"))
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_image(self, index: int) -> bool:
        """Set the image by index."""
        await self._get_current_image_index()
        total_images = self.get_total_images()

        if index < 0 or index >= total_images:
            raise ValueError(
                f"Image index {index} is out of range. Total images: {total_images}."
            )

        all_images_index = self.get_all_images_index()
        img_index = all_images_index[index]
        result = await self._send_command(COMMAND_SET_IMAGE.format(f"{img_index:02X}"))
        return self._check_command_result(result, 0, {1})

    def get_all_images_index(self) -> list[int] | None:
        """Return cached list of all image indexes."""
        return self._get_adv_value("all_images_index")

    def get_current_image_index(self) -> int | None:
        """Return cached current image index."""
        return self._get_adv_value("current_image_index")

    def get_total_images(self) -> int | None:
        """Return cached total number of images."""
        return self._get_adv_value("total_num_of_images")



"""Air Purifier adv parser."""

from __future__ import annotations

import struct

from ..const.air_purifier import AirPurifierMode, AirQualityLevel


def process_air_purifier(
    data: bytes | None, mfr_data: bytes | None
) -> dict[str, bool | int]:
    """Process air purifier services data."""
    if mfr_data is None or len(mfr_data) < 14:
        return {}
    device_data = mfr_data[6:]

    _seq_num = device_data[0]
    _isOn = bool(device_data[1] & 0b10000000)
    _mode = device_data[1] & 0b00000111
    _is_aqi_valid = bool(device_data[2] & 0b00000100)
    _child_lock = bool(device_data[2] & 0b00000010)
    _speed = device_data[3] & 0b01111111
    _aqi_level = (device_data[4] & 0b00000110) >> 1
    _aqi_level = AirQualityLevel(_aqi_level).name.lower()
    _work_time = struct.unpack(">H", device_data[5:7])[0]
    _err_code = device_data[7]

    return {
        "isOn": _isOn,
        "mode": get_air_purifier_mode(_mode, _speed),
        "isAqiValid": _is_aqi_valid,
        "child_lock": _child_lock,
        "speed": _speed,
        "aqi_level": _aqi_level,
        "filter element working time": _work_time,
        "err_code": _err_code,
        "sequence_number": _seq_num,
    }


def get_air_purifier_mode(mode: int, speed: int) -> str | None:
    if mode == 1:
        if 0 <= speed <= 33:
            return "level_1"
        if 34 <= speed <= 66:
            return "level_2"
        return "level_3"
    if 1 < mode <= 4:
        mode += 2
        return AirPurifierMode(mode).name.lower()
    return None



"""Meter parser."""

from __future__ import annotations

import struct
from typing import Any

from ._sensor_th import decode_temp_humidity

CO2_UNPACK = struct.Struct(">H").unpack_from

# Meter Pro CO2 sensor spec range is 400-9999 ppm. Higher values are
# transient parsing artifacts and surface as huge spikes downstream.
CO2_MAX_PPM = 9999


def process_wosensorth(data: bytes | None, mfr_data: bytes | None) -> dict[str, Any]:
    """Process woSensorTH/Temp sensor services data."""
    temp_data: bytes | None = None
    battery: int | None = None

    if mfr_data and len(mfr_data) >= 11:
        temp_data = mfr_data[8:11]

    if data and len(data) >= 3:
        if not temp_data and len(data) >= 6:
            temp_data = data[3:6]
        battery = data[2] & 0b01111111

    if not temp_data:
        return {}

    return decode_temp_humidity(temp_data, battery)


def process_wosensorth_c(data: bytes | None, mfr_data: bytes | None) -> dict[str, Any]:
    """Process woSensorTH/Temp sensor services data with CO2."""
    _wosensorth_data = process_wosensorth(data, mfr_data)
    if _wosensorth_data and mfr_data and len(mfr_data) >= 15:
        co2_data = mfr_data[13:15]
        co2 = CO2_UNPACK(co2_data)[0]
        if co2 <= CO2_MAX_PPM:
            _wosensorth_data["co2"] = co2
    return _wosensorth_data



import logging
from typing import Any

from ..adv_parsers.humidifier import calculate_temperature_and_humidity
from ..const import SwitchbotModel
from ..const.evaporative_humidifier import (
    TARGET_HUMIDITY_MODES,
    HumidifierAction,
    HumidifierMode,
    HumidifierWaterLevel,
)
from .device import (
    SwitchbotEncryptedDevice,
    SwitchbotOperationError,
    SwitchbotSequenceDevice,
    update_after_operation,
)

_LOGGER = logging.getLogger(__name__)

COMMAND_HEADER = "57"
COMMAND_GET_CK_IV = f"{COMMAND_HEADER}0f2103"
COMMAND_TURN_ON = f"{COMMAND_HEADER}0f430101"
COMMAND_CHILD_LOCK_ON = f"{COMMAND_HEADER}0f430501"
COMMAND_CHILD_LOCK_OFF = f"{COMMAND_HEADER}0f430500"
COMMAND_AUTO_DRY_ON = f"{COMMAND_HEADER}0f430a01"
COMMAND_AUTO_DRY_OFF = f"{COMMAND_HEADER}0f430a02"
COMMAND_SET_MODE = f"{COMMAND_HEADER}0f4302"
COMMAND_GET_BASIC_INFO = f"{COMMAND_HEADER}000300"
COMMAND_SET_DRYING_FILTER = f"{COMMAND_TURN_ON}08"

MODES_COMMANDS = {
    HumidifierMode.HIGH: "010100",
    HumidifierMode.MEDIUM: "010200",
    HumidifierMode.LOW: "010300",
    HumidifierMode.QUIET: "010400",
    HumidifierMode.TARGET_HUMIDITY: "0200",
    HumidifierMode.SLEEP: "0300",
    HumidifierMode.AUTO: "040000",
}

DEVICE_GET_BASIC_SETTINGS_KEY = "570f4481"


class SwitchbotEvaporativeHumidifier(SwitchbotSequenceDevice, SwitchbotEncryptedDevice):
    """Representation of a Switchbot Evaporative Humidifier"""

    _model = SwitchbotModel.EVAPORATIVE_HUMIDIFIER
    _turn_on_command = COMMAND_TURN_ON
    _turn_off_command = f"{COMMAND_HEADER}0f430100"
    _force_next_update: bool = False

    async def get_basic_info(self) -> dict[str, Any] | None:
        """Get device basic settings."""
        if not (_data := await self._get_basic_info(DEVICE_GET_BASIC_SETTINGS_KEY)):
            return None

        _LOGGER.debug("basic info data: %s", _data.hex())
        isOn = bool(_data[1] & 0b10000000)
        mode = HumidifierMode(_data[1] & 0b00001111)
        over_humidify_protection = bool(_data[2] & 0b10000000)
        child_lock = bool(_data[2] & 0b00100000)
        tank_removed = bool(_data[2] & 0b00000100)
        tilted_alert = bool(_data[2] & 0b00000010)
        filter_missing = bool(_data[2] & 0b00000001)
        is_meter_binded = bool(_data[3] & 0b10000000)

        _temp_c, _temp_f, humidity = calculate_temperature_and_humidity(
            _data[3:6], is_meter_binded
        )

        water_level = HumidifierWaterLevel(_data[5] & 0b00000011).name.lower()
        filter_run_time = int.from_bytes(_data[6:8], byteorder="big") & 0xFFF
        target_humidity = _data[10] & 0b01111111

        return {
            "isOn": isOn,
            "mode": mode,
            "over_humidify_protection": over_humidify_protection,
            "child_lock": child_lock,
            "tank_removed": tank_removed,
            "tilted_alert": tilted_alert,
            "filter_missing": filter_missing,
            "is_meter_binded": is_meter_binded,
            "humidity": humidity,
            "temperature": _temp_c,
            "temp": {"c": _temp_c, "f": _temp_f},
            "water_level": water_level,
            "filter_run_time": filter_run_time,
            "target_humidity": target_humidity,
        }

    @update_after_operation
    async def set_target_humidity(self, target_humidity: int) -> bool:
        """Set target humidity."""
        self._validate_water_level()
        self._validate_mode_for_target_humidity()
        command = (
            COMMAND_SET_MODE
            + MODES_COMMANDS[self.get_mode()]
            + f"{target_humidity:02x}"
        )
        result = await self._send_command(command)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_mode(self, mode: HumidifierMode) -> bool:
        """Set device mode."""
        self._validate_water_level()
        self._validate_meter_binding(mode)

        if mode == HumidifierMode.DRYING_FILTER:
            command = COMMAND_SET_DRYING_FILTER
        else:
            command = COMMAND_SET_MODE + MODES_COMMANDS[mode]

        if mode in TARGET_HUMIDITY_MODES:
            target_humidity = self.get_target_humidity()
            if target_humidity is None:
                raise SwitchbotOperationError(
                    "Target humidity must be set before switching to target humidity mode or sleep mode"
                )
            command += f"{target_humidity:02x}"
        result = await self._send_command(command)
        return self._check_command_result(result, 0, {1})

    def _validate_water_level(self, mode: HumidifierMode | None = None) -> None:
        """Validate that the water level is not empty."""
        if mode == HumidifierMode.DRYING_FILTER:
            return

        if self.get_water_level() == HumidifierWaterLevel.EMPTY.name.lower():
            raise SwitchbotOperationError(
                "Cannot perform operation when water tank is empty"
            )

    def _validate_mode_for_target_humidity(self) -> None:
        """Validate that the current mode supports target humidity."""
        if self.get_mode() not in TARGET_HUMIDITY_MODES:
            raise SwitchbotOperationError(
                "Target humidity can only be set in target humidity mode or sleep mode"
            )

    def _validate_meter_binding(self, mode: HumidifierMode) -> None:
        """Validate that the meter is binded for specific modes."""
        if not self.is_meter_binded() and mode in [
            HumidifierMode.TARGET_HUMIDITY,
            HumidifierMode.AUTO,
        ]:
            raise SwitchbotOperationError(
                "Cannot set target humidity or auto mode when meter is not binded"
            )

    @update_after_operation
    async def set_child_lock(self, enabled: bool) -> bool:
        """Set child lock."""
        result = await self._send_command(
            COMMAND_CHILD_LOCK_ON if enabled else COMMAND_CHILD_LOCK_OFF
        )
        return self._check_command_result(result, 0, {1})

    def is_on(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("isOn")

    def get_mode(self) -> HumidifierMode | None:
        """Return state from cache."""
        return self._get_adv_value("mode")

    def is_child_lock_enabled(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("child_lock")

    def is_over_humidify_protection_enabled(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("over_humidify_protection")

    def is_tank_removed(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("tank_removed")

    def is_filter_missing(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("filter_missing")

    def is_filter_alert_on(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("filter_alert")

    def is_tilted_alert_on(self) -> bool | None:
        """Return state from cache."""
        return self._get_adv_value("tilted_alert")

    def get_water_level(self) -> HumidifierWaterLevel | None:
        """Return state from cache."""
        return self._get_adv_value("water_level")

    def get_filter_run_time(self) -> int | None:
        """Return state from cache."""
        return self._get_adv_value("filter_run_time")

    def get_target_humidity(self) -> int | None:
        """Return state from cache."""
        return self._get_adv_value("target_humidity")

    def get_humidity(self) -> int | None:
        """Return state from cache."""
        return self._get_adv_value("humidity")

    def get_temperature(self) -> float | None:
        """Return state from cache."""
        return self._get_adv_value("temperature")

    def get_action(self) -> int:
        """Return current action from cache."""
        if not self.is_on():
            return HumidifierAction.OFF
        if self.get_mode() != HumidifierMode.DRYING_FILTER:
            return HumidifierAction.HUMIDIFYING
        return HumidifierAction.DRYING

    def is_meter_binded(self) -> bool | None:
        """Return meter bind state from cache."""
        return self._get_adv_value("is_meter_binded")
